# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Conftest for google.genai tests."""

import datetime
import os
from unittest import mock
import uuid
import pytest

from .. import _common
from .. import _replay_api_client
from .. import client as google_genai_client_module


def pytest_addoption(parser):
  parser.addoption(
      '--mode',
      action='store',
      default='auto',
      help="""Replay mode.
  One of:
  * auto: Replay if replay files exist, otherwise record.
  * record: Always call the API and record.
  * replay: Always replay, fail if replay files do not exist.
  * api: Always call the API and do not record.
  * tap: Always replay, fail if replay files do not exist. Also sets default values for the API key and replay directory.
""",
  )
  parser.addoption(
      '--private',
      action='store_true',
      default=False,
      help='Run private tests.',
  )


# Overridden via parameterized test.
@pytest.fixture
def use_vertex():
  return False


# Overridden at the module level for each test file.
@pytest.fixture
def replays_prefix():
  return 'test'


def _get_replay_id(use_vertex: bool, replays_prefix: str) -> str:
  test_name_ending = os.environ.get('PYTEST_CURRENT_TEST').split('::')[-1]
  test_name = (
      test_name_ending.split(' ')[0].split('[')[0]
      + '.'
      + ('vertex' if use_vertex else 'mldev')
  )
  return '/'.join([replays_prefix, test_name])


@pytest.fixture
def client(use_vertex, replays_prefix, http_options, request):
  mode = request.config.getoption('--mode')
  if mode not in ['auto', 'record', 'replay', 'api', 'tap']:
    raise ValueError('Invalid mode: ' + mode)
  test_function_name = request.function.__name__
  test_filename = os.path.splitext(os.path.basename(request.path))[0]
  if test_function_name.startswith(test_filename):
    raise ValueError(f"""
      {test_function_name}:
      Do not include the test filename in the test function name.
      keep the test function name short.""")
  for word in ['mldev', 'ml_dev', 'vertex']:
    if word in test_function_name:
      raise ValueError(f"""
        {test_function_name}:
        You should never ever include api types in the file name (mldev or vertex).
        All tests should run against all apis.
        Assert an exception if the test is not supported in an API.""")
  replay_id = _get_replay_id(use_vertex, replays_prefix)

  if mode == 'tap':
    mode = 'replay'

    # Set various environment variables to ensure that the test runs.
    os.environ['GOOGLE_API_KEY'] = 'dummy-api-key'
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = os.path.join(
        os.path.dirname(__file__),
        'credentials.json',
    )
    os.environ['GOOGLE_CLOUD_PROJECT'] = 'project-id'
    os.environ['GOOGLE_CLOUD_LOCATION'] = 'location'

    # Set the replay directory to the root directory of the replays.
    # This is needed to ensure that the replay files are found.
    replays_root_directory = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '../../../../../google/cloud/aiplatform/sdk/genai/replays',
        )
    )
    os.environ['GOOGLE_GENAI_REPLAYS_DIRECTORY'] = replays_root_directory
  # Get private arg.
  private = request.config.getoption('--private')
  replay_client = _replay_api_client.ReplayApiClient(
      mode=mode,
      replay_id=replay_id,
      vertexai=use_vertex,
      http_options=http_options,
      private=private,
  )

  with mock.patch.object(
      google_genai_client_module.Client, '_get_api_client'
  ) as patch_method:
    patch_method.return_value = replay_client
    google_genai_client = google_genai_client_module.Client(vertexai=use_vertex)

    # Yield the client so that cleanup can be completed at the end of the test.
    yield google_genai_client

    # Save the replay after the test if we were in recording mode.
    google_genai_client._api_client.close()


@pytest.fixture
def mock_timestamped_unique_name():
  with mock.patch.object(
      _common,
      'timestamped_unique_name',
      return_value='20240101000000_bd656',
  ) as unique_name_mock:
    yield unique_name_mock


@pytest.fixture
def image_jpeg():
  import PIL.Image
  image_path = os.path.join(os.path.dirname(__file__), 'data', 'google.jpg')
  with PIL.Image.open(image_path) as img:
    yield img


@pytest.fixture
def image_png():
  import PIL.Image
  image_path = os.path.join(os.path.dirname(__file__), 'data', 'google.png')
  with PIL.Image.open(image_path) as img:
    yield img



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from typing_extensions import TypeAlias

from .text_content_param import TextContentParam
from .audio_content_param import AudioContentParam
from .image_content_param import ImageContentParam
from .video_content_param import VideoContentParam
from .thought_content_param import ThoughtContentParam
from .document_content_param import DocumentContentParam
from .function_call_content_param import FunctionCallContentParam
from .function_result_content_param import FunctionResultContentParam
from .file_search_call_content_param import FileSearchCallContentParam
from .url_context_call_content_param import URLContextCallContentParam
from .file_search_result_content_param import FileSearchResultContentParam
from .google_search_call_content_param import GoogleSearchCallContentParam
from .url_context_result_content_param import URLContextResultContentParam
from .code_execution_call_content_param import CodeExecutionCallContentParam
from .google_search_result_content_param import GoogleSearchResultContentParam
from .mcp_server_tool_call_content_param import MCPServerToolCallContentParam
from .code_execution_result_content_param import CodeExecutionResultContentParam
from .mcp_server_tool_result_content_param import MCPServerToolResultContentParam

__all__ = ["ContentParam"]

ContentParam: TypeAlias = Union[
    TextContentParam,
    ImageContentParam,
    AudioContentParam,
    DocumentContentParam,
    VideoContentParam,
    ThoughtContentParam,
    FunctionCallContentParam,
    FunctionResultContentParam,
    CodeExecutionCallContentParam,
    CodeExecutionResultContentParam,
    URLContextCallContentParam,
    URLContextResultContentParam,
    GoogleSearchCallContentParam,
    GoogleSearchResultContentParam,
    MCPServerToolCallContentParam,
    MCPServerToolResultContentParam,
    FileSearchCallContentParam,
    FileSearchResultContentParam,
]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from copy import deepcopy
import datetime
import pytest
import sys
from ... import types
from .. import pytest_helper
from ... import _transformers as t


_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI = types._CreateCachedContentParameters(
    model='gemini-2.5-flash',
    config={
        'contents': [
            types.Content(
                role='user',
                parts=[
                    types.Part(
                        fileData=types.FileData(
                            fileUri='gs://cloud-samples-data/generative-ai/pdf/2312.11805v3.pdf',
                            mimeType='application/pdf',
                        )
                    ),
                    types.Part(
                        fileData=types.FileData(
                            fileUri='gs://cloud-samples-data/generative-ai/pdf/2403.05530.pdf',
                            mimeType='application/pdf',
                        )
                    ),
                ],
            )
        ],
        'system_instruction': t.t_content('What is the sum of the two pdfs?'),
        'display_name': 'test cache',
        'ttl': '86400s',
    },
)

_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE = types._CreateCachedContentParameters(
    model='gemini-2.5-flash',
    config={
        'contents': [
            types.Content(
                role='user',
                parts=[
                    types.Part(
                        fileData=types.FileData(
                            mimeType='application/pdf',
                            fileUri='https://generativelanguage.googleapis.com/v1beta/files/v200dhvn15h7',
                        )
                    )
                ],
            )
        ],
        'system_instruction': t.t_content('What is the sum of the two pdfs?'),
        'display_name': 'test cache',
        'ttl': '86400s',
    },
)

_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_PARTIAL_MODEL_1 = deepcopy(
    _CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI
)
_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_PARTIAL_MODEL_1.model = (
    'models/gemini-2.5-flash'
)
_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_PARTIAL_MODEL_2 = deepcopy(
    _CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI
)
_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_PARTIAL_MODEL_2.model = (
    'publishers/google/models/gemini-2.5-flash'
)

_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE_PARTIAL_MODEL_1 = deepcopy(
    _CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE
)
_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE_PARTIAL_MODEL_1.model = (
    'models/gemini-2.5-flash'
)

_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_CMEK = deepcopy(
    _CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI
)
_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_CMEK.config.kms_key_name = (
    'projects/test-project/locations/us-central1/keyRings/test-keyring/cryptoKeys/test-key'
)

if sys.version_info >= (3, 11):
  _EXPIRE_TIME = datetime.datetime.fromisoformat('2025-12-20T00:00:00Z')
else:
  _EXPIRE_TIME = datetime.datetime.fromisoformat('2025-12-20T00:00:00+00:00')

_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_EXPIRE_TIME = deepcopy(
    _CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI
)
_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_EXPIRE_TIME.config.ttl = None
_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_EXPIRE_TIME.config.expire_time = (
    _EXPIRE_TIME
)
_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_EXPIRE_TIME.config.display_name = (
    'test cache'
)

_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE_EXPIRE_TIME = deepcopy(
    _CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE
)
_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE_EXPIRE_TIME.config.ttl = None
_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE_EXPIRE_TIME.config.expire_time = (
    _EXPIRE_TIME
)
_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE_EXPIRE_TIME.config.display_name = (
    'test cache'
)


# Replay mode is not supported for caches tests due to the error message
# inconsistency in api and replay mode.
# To run api mode tests, use the following steps:
# 1. First create the resource.
#   sh run_tests.sh pytest -s tests/caches/test_create.py --mode=api
#   1.1 If mldev test_create fails, update the uploaded file using this colab
#       https://colab.sandbox.google.com/drive/1Fv6KGSs0cg6tlpcUHdsclHussXMEGOXk#scrollTo=RSKmFPx00MVL.
# 2. Find the resource name in debugging print and change the resource name constants.py.
# 3. Run and record get and update tests.
#   sh run_tests.sh pytest -s tests/caches/test_get.py --mode=api && sh run_tests.sh pytest -s tests/caches/test_update.py --mode=api && sh run_tests.sh pytest -s tests/caches/test_delete.py --mode=api
test_table: list[pytest_helper.TestTableItem] = [
    pytest_helper.TestTableItem(
        name='test_caches_create_with_gcs_uri',
        exception_if_mldev='INVALID_ARGUMENT',
        parameters=_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI,
    ),
    pytest_helper.TestTableItem(
        name='test_caches_create_with_gcs_uri_cmek',
        exception_if_mldev='not supported',
        exception_if_vertex='INVALID_ARGUMENT',  # The key is invalid.
        parameters=_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_CMEK,
    ),
    pytest_helper.TestTableItem(
        name='test_caches_create_with_gcs_uri_expire_time',
        exception_if_mldev='INVALID_ARGUMENT',
        parameters=_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_EXPIRE_TIME,
    ),
    pytest_helper.TestTableItem(
        name='test_caches_create_model_partial_path_1',
        exception_if_mldev='INVALID_ARGUMENT',
        parameters=_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_PARTIAL_MODEL_1,
    ),
    pytest_helper.TestTableItem(
        name='test_caches_create_model_partial_path_2',
        exception_if_mldev='404',
        parameters=_CREATE_CACHED_CONTENT_PARAMETERS_GCS_URI_PARTIAL_MODEL_2,
    ),
    pytest_helper.TestTableItem(
        name='test_caches_create_with_googleai_file',
        exception_if_vertex='Internal',
        parameters=_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE,
        skip_in_api_mode='Create is not reproducible in the API mode.',
    ),
    pytest_helper.TestTableItem(
        name='test_caches_create_with_googleai_file_expire_time',
        exception_if_vertex='Internal',
        parameters=_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE_EXPIRE_TIME,
        skip_in_api_mode='Create is not reproducible in the API mode.',
    ),
    pytest_helper.TestTableItem(
        name='test_caches_create_with_googleai_file_model_partial_path_1',
        exception_if_vertex='Internal',
        parameters=_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE_PARTIAL_MODEL_1,
        skip_in_api_mode='Create is not reproducible in the API mode.',
    ),
]
pytestmark = [
    pytest_helper.setup(
        file=__file__,
        globals_for_file=globals(),
        test_method='caches.create',
        test_table=test_table,
    ),
]
pytest_plugins = ('pytest_asyncio',)


@pytest.mark.asyncio
async def test_async_googleai_file_create(client):
  if client._api_client.vertexai:
    with pytest.raises(Exception):
      await client.aio.caches.create(
          model=_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE.model,
          config=_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE.config,
      )
  else:
    await client.aio.caches.create(
        model=_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE.model,
        config=_CREATE_CACHED_CONTENT_PARAMETERS_GOOGLEAI_FILE.config,
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from unittest.mock import MagicMock, mock_open, patch

import sentencepiece as spm
from sentencepiece import sentencepiece_model_pb2

from ... import _local_tokenizer_loader as loader

# A minimal valid sentencepiece model proto
FAKE_MODEL_CONTENT = sentencepiece_model_pb2.ModelProto(
    pieces=[
        sentencepiece_model_pb2.ModelProto.SentencePiece(
            piece="<unk>",
            score=0,
            type=sentencepiece_model_pb2.ModelProto.SentencePiece.Type.UNKNOWN,
        ),
        sentencepiece_model_pb2.ModelProto.SentencePiece(
            piece="<s>",
            score=0,
            type=sentencepiece_model_pb2.ModelProto.SentencePiece.Type.CONTROL,
        ),
        sentencepiece_model_pb2.ModelProto.SentencePiece(
            piece="</s>",
            score=0,
            type=sentencepiece_model_pb2.ModelProto.SentencePiece.Type.CONTROL,
        ),
        sentencepiece_model_pb2.ModelProto.SentencePiece(
            piece="a",
            score=0,
            type=sentencepiece_model_pb2.ModelProto.SentencePiece.Type.NORMAL,
        ),
    ]
).SerializeToString()

GEMMA2_HASH = "61a7b147390c64585d6c3543dd6fc636906c9af3865a5548f27f31aee1d4c8e2"


class TestGetTokenizerName(unittest.TestCase):

  def test_get_tokenizer_name_success(self):
    self.assertEqual(loader.get_tokenizer_name("gemini-2.5-pro"), "gemma3")
    self.assertEqual(
        loader.get_tokenizer_name("gemini-2.5-pro-preview-06-05"), "gemma3"
    )

  def test_get_tokenizer_name_unsupported(self):
    with self.assertRaisesRegex(
        ValueError, "Model unsupported-model is not supported"
    ):
      loader.get_tokenizer_name("unsupported-model")


@patch("genai._local_tokenizer_loader.os.rename")
@patch("genai._local_tokenizer_loader.os.makedirs")
@patch("genai._local_tokenizer_loader.os.remove")
@patch("genai._local_tokenizer_loader.open", new_callable=mock_open)
@patch("genai._local_tokenizer_loader.os.path.exists")
@patch("genai._local_tokenizer_loader.requests.get")
@patch("genai._local_tokenizer_loader.hashlib.sha256")
class TestLoaderFunctions(unittest.TestCase):

  def setUp(self):
    # Clear caches before each test
    loader.load_model_proto.cache_clear()
    loader.get_sentencepiece.cache_clear()
    # Patch tempfile.gettempdir to control cache location
    self.tempdir_patcher = patch(
        "tempfile.gettempdir", return_value="/tmp/fake_temp_dir"
    )
    self.mock_tempdir = self.tempdir_patcher.start()

  def tearDown(self):
    self.tempdir_patcher.stop()

  def _setup_get_mock(self, mock_get):
    mock_response = MagicMock()
    mock_response.content = FAKE_MODEL_CONTENT
    mock_response.raise_for_status.return_value = None
    mock_get.return_value = mock_response

  def test_load_model_proto_from_url(
      self,
      mock_sha256,
      mock_get,
      mock_exists,
      mock_open_func,
      mock_remove,
      mock_makedirs,
      mock_rename,
  ):
    mock_exists.return_value = False  # Don't use cache
    self._setup_get_mock(mock_get)
    mock_sha256.return_value.hexdigest.return_value = GEMMA2_HASH

    proto = loader.load_model_proto("gemma2")

    self.assertIsInstance(proto, sentencepiece_model_pb2.ModelProto)
    self.assertEqual(len(proto.pieces), 4)
    mock_get.assert_called_once()
    mock_makedirs.assert_called_once()
    mock_open_func.assert_called()
    mock_rename.assert_called_once()

  def test_load_model_proto_from_cache(
      self,
      mock_sha256,
      mock_get,
      mock_exists,
      mock_open_func,
      mock_remove,
      mock_makedirs,
      mock_rename,
  ):
    mock_exists.return_value = True  # Use cache
    mock_open_func.return_value.read.return_value = FAKE_MODEL_CONTENT
    mock_sha256.return_value.hexdigest.return_value = GEMMA2_HASH

    proto = loader.load_model_proto("gemma2")

    self.assertIsInstance(proto, sentencepiece_model_pb2.ModelProto)
    mock_get.assert_not_called()

  def test_load_model_proto_corrupted_cache(
      self,
      mock_sha256,
      mock_get,
      mock_exists,
      mock_open_func,
      mock_remove,
      mock_makedirs,
      mock_rename,
  ):
    mock_exists.return_value = True  # Use cache initially
    self._setup_get_mock(mock_get)
    mock_open_func.return_value.__enter__.return_value.read.return_value = (
        b"corrupted"
    )

    # First hash for corrupted cache, second for good download
    mock_sha256.side_effect = [
        MagicMock(hexdigest=MagicMock(return_value="wrong_hash")),
        MagicMock(hexdigest=MagicMock(return_value=GEMMA2_HASH)),
    ]

    proto = loader.load_model_proto("gemma2")

    self.assertIsInstance(proto, sentencepiece_model_pb2.ModelProto)
    mock_remove.assert_called_once()
    mock_get.assert_called_once()

  def test_load_model_proto_bad_hash_from_url(
      self,
      mock_sha256,
      mock_get,
      mock_exists,
      mock_open_func,
      mock_remove,
      mock_makedirs,
      mock_rename,
  ):
    mock_exists.return_value = False
    self._setup_get_mock(mock_get)
    mock_sha256.return_value.hexdigest.return_value = "wrong_hash"

    with self.assertRaisesRegex(
        ValueError, "Downloaded model file is corrupted"
    ):
      loader.load_model_proto("gemma2")

  def test_load_model_proto_unsupported(self, *args):
    with self.assertRaisesRegex(
        ValueError, "Tokenizer unsupported is not supported"
    ):
      loader.load_model_proto("unsupported")

  def test_get_sentencepiece_success(
      self,
      mock_sha256,
      mock_get,
      mock_exists,
      mock_open_func,
      mock_remove,
      mock_makedirs,
      mock_rename,
  ):
    mock_exists.return_value = False
    self._setup_get_mock(mock_get)
    mock_sha256.return_value.hexdigest.return_value = GEMMA2_HASH

    processor = loader.get_sentencepiece("gemma2")

    self.assertIsInstance(processor, spm.SentencePieceProcessor)
    mock_get.assert_called_once()

  def test_get_sentencepiece_unsupported(self, *args):
    with self.assertRaisesRegex(
        ValueError, "Tokenizer unsupported is not supported"
    ):
      loader.get_sentencepiece("unsupported")

  def test_get_sentencepiece_caching(
      self,
      mock_sha256,
      mock_get,
      mock_exists,
      mock_open_func,
      mock_remove,
      mock_makedirs,
      mock_rename,
  ):
    mock_exists.return_value = False
    self._setup_get_mock(mock_get)
    mock_sha256.return_value.hexdigest.return_value = GEMMA2_HASH

    # Call twice
    loader.get_sentencepiece("gemma2")
    loader.get_sentencepiece("gemma2")

    # Should only be loaded once due to lru_cache
    mock_get.assert_called_once()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests for batches.create() with BigQuery source."""

import pytest

from ... import types
from .. import pytest_helper


_GEMINI_MODEL = 'gemini-2.5-flash'
_GEMINI_MODEL_FULL_NAME = 'publishers/google/models/gemini-2.5-flash'
_EMBEDDING_MODEL = 'gemini-embedding-001'
_EMBEDDING_MODEL_FULL_NAME = 'publishers/google/models/gemini-embedding-001'
_DISPLAY_NAME = 'test_batch'
_GENERATE_CONTENT_BQ_INPUT_FILE = (
    'bq://vertex-sdk-dev.unified_genai_tests_batches.generate_content_requests'
)
_GENERATE_CONTENT_BQ_OUTPUT_PREFIX = (
    'bq://vertex-sdk-dev.unified_genai_tests_batches.generate_content_output'
)

_EMBEDDING_BQ_INPUT_FILE = (
    'bq://vertex-sdk-dev.unified_genai_tests_batches.embedding_requests'
)
_EMBEDDING_BQ_OUTPUT_PREFIX = (
    'bq://vertex-sdk-dev.unified_genai_tests_batches.embedding_output'
)


# All tests will be run for both Vertex and MLDev.
test_table: list[pytest_helper.TestTableItem] = [
    pytest_helper.TestTableItem(
        name='test_union_generate_content_with_bigquery',
        parameters=types._CreateBatchJobParameters(
            model=_GEMINI_MODEL_FULL_NAME,
            src=_GENERATE_CONTENT_BQ_INPUT_FILE,
        ),
        exception_if_mldev='not supported in Gemini API',
        has_union=True,
    ),
    pytest_helper.TestTableItem(
        name='test_union_embedding_with_bigquery',
        parameters=types._CreateBatchJobParameters(
            model='publishers/google/models/text-embedding-004',
            src=_EMBEDDING_BQ_INPUT_FILE,
            config={
                'display_name': _DISPLAY_NAME,
                'dest': _EMBEDDING_BQ_OUTPUT_PREFIX,
            },
        ),
        exception_if_mldev='not supported in Gemini API',
        has_union=True,
    ),
    pytest_helper.TestTableItem(
        name='test_generate_content_with_bigquery',
        parameters=types._CreateBatchJobParameters(
            model=_GEMINI_MODEL_FULL_NAME,
            src={
                'bigquery_uri': _GENERATE_CONTENT_BQ_INPUT_FILE,
                'format': 'bigquery',
            },
            config={
                'display_name': _DISPLAY_NAME,
                'dest': {
                    'bigquery_uri': _GENERATE_CONTENT_BQ_OUTPUT_PREFIX,
                    'format': 'bigquery',
                },
            },
        ),
        exception_if_mldev='one of',
    ),
]

pytestmark = [
    pytest.mark.usefixtures('mock_timestamped_unique_name'),
    pytest_helper.setup(
        file=__file__,
        globals_for_file=globals(),
        test_method='batches.create',
        test_table=test_table,
    ),
]


@pytest.mark.asyncio
async def test_async_create(client):
  with pytest_helper.exception_if_mldev(client, ValueError):
    batch_job = await client.aio.batches.create(
        model=_GEMINI_MODEL,
        src=_GENERATE_CONTENT_BQ_INPUT_FILE,
    )

    assert batch_job.name.startswith('projects/')
    assert (
        batch_job.model == _GEMINI_MODEL_FULL_NAME
    )  # Converted to Vertex full name.
    assert batch_job.src.bigquery_uri == _GENERATE_CONTENT_BQ_INPUT_FILE
    assert batch_job.src.format == 'bigquery'



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from unittest.mock import MagicMock, patch

from sentencepiece import sentencepiece_model_pb2

from ... import local_tokenizer
from ... import types


class TestLocalTokenizer(unittest.TestCase):

  def setUp(self):
    # This setup will be used by all tests
    self.mock_load_model_proto = patch(
        'genai._local_tokenizer_loader.load_model_proto'
    ).start()
    self.mock_get_sentencepiece = patch(
        'genai._local_tokenizer_loader.get_sentencepiece'
    ).start()

    self.mock_load_model_proto.return_value = MagicMock()
    self.mock_tokenizer = MagicMock()
    self.mock_get_sentencepiece.return_value = self.mock_tokenizer

    self.tokenizer = local_tokenizer.LocalTokenizer(model_name='gemini-3-pro-preview')

  def tearDown(self):
    patch.stopall()

  def test_count_tokens_simple_string(self):
    self.mock_tokenizer.encode.return_value = [[1, 2, 3]]
    result = self.tokenizer.count_tokens('Hello world')
    self.assertEqual(result.total_tokens, 3)
    self.mock_tokenizer.encode.assert_called_once_with(['Hello world'])

  def test_count_tokens_list_of_strings(self):
    self.mock_tokenizer.encode.return_value = [[1, 2], [3]]
    result = self.tokenizer.count_tokens(['Hello', 'world'])
    self.assertEqual(result.total_tokens, 3)
    self.mock_tokenizer.encode.assert_called_once_with(['Hello', 'world'])

  def test_count_tokens_with_content_object(self):
    self.mock_tokenizer.encode.return_value = [[1, 2, 3]]
    content = types.Content(parts=[types.Part(text='Hello world')])
    result = self.tokenizer.count_tokens(content)
    self.assertEqual(result.total_tokens, 3)
    self.mock_tokenizer.encode.assert_called_once_with(['Hello world'])

  def test_count_tokens_with_chat_history(self):
    self.mock_tokenizer.encode.return_value = [[1, 2], [3, 4, 5]]
    history = [
        types.Content(role='user', parts=[types.Part(text='Hello')]),
        types.Content(role='model', parts=[types.Part(text='Hi there!')]),
    ]
    result = self.tokenizer.count_tokens(history)
    self.assertEqual(result.total_tokens, 5)
    self.mock_tokenizer.encode.assert_called_once_with(['Hello', 'Hi there!'])

  def test_count_tokens_with_tools(self):
    self.mock_tokenizer.encode.return_value = [
        [1],
        [1, 2],
        [1, 2, 3],
        [1, 2, 3, 4],
        [1, 2, 3, 4, 5],
        [1, 2, 3, 4, 5, 6],
    ]
    tool = types.Tool(
        function_declarations=[
            types.FunctionDeclaration(
                name='get_weather',
                description='Get the weather for a location',
                parameters=types.Schema(
                    type=types.Type.OBJECT,
                    properties={
                        'location': types.Schema(
                            type=types.Type.STRING, description='The location'
                        )
                    },
                    required=['location'],
                ),
            )
        ]
    )
    config = types.CountTokensConfig(tools=[tool])
    result = self.tokenizer.count_tokens(
        'What is the weather in Boston?', config=config
    )
    self.assertEqual(result.total_tokens, 21)
    self.mock_tokenizer.encode.assert_called_once_with([
        'What is the weather in Boston?',
        'get_weather',
        'Get the weather for a location',
        'location',
        'location',
        'The location',
    ])

  def test_count_tokens_with_function_call(self):
    self.mock_tokenizer.encode.return_value = [[1, 2], [3], [4, 5]]
    content = types.Content(
        role='model',
        parts=[
            types.Part(
                function_call=types.FunctionCall(
                    name='get_weather', args={'location': 'Boston'}
                )
            )
        ],
    )
    result = self.tokenizer.count_tokens(content)
    self.assertEqual(result.total_tokens, 5)
    self.mock_tokenizer.encode.assert_called_once_with(
        ['get_weather', 'location', 'Boston']
    )

  def test_count_tokens_with_function_response(self):
    self.mock_tokenizer.encode.return_value = [[1, 2], [3], [4, 5]]
    content = types.Content(
        role='user',
        parts=[
            types.Part(
                function_response=types.FunctionResponse(
                    name='get_weather', response={'weather': 'sunny'}
                )
            )
        ],
    )
    result = self.tokenizer.count_tokens(content)
    self.assertEqual(result.total_tokens, 5)
    self.mock_tokenizer.encode.assert_called_once_with(
        ['get_weather', 'weather', 'sunny']
    )

  def test_count_tokens_with_unsupported_content(self):
    with self.assertRaises(ValueError):
      self.tokenizer.count_tokens(
          [
              types.Content(
                  parts=[
                      types.Part(
                          inline_data=types.Blob(
                              data=b'test', mime_type='image/png'
                          )
                      )
                  ]
              )
          ]
      )

  def test_count_tokens_with_system_instruction(self):
    self.mock_tokenizer.encode.return_value = [[1, 2, 3], [4, 5]]
    config = types.CountTokensConfig(
        system_instruction=types.Content(
            parts=[types.Part(text='You are a helpful assistant.')]
        )
    )
    result = self.tokenizer.count_tokens('Hello', config=config)
    self.assertEqual(result.total_tokens, 5)
    self.mock_tokenizer.encode.assert_called_once_with(
        ['Hello', 'You are a helpful assistant.']
    )

  def test_count_tokens_with_response_schema(self):
    self.mock_tokenizer.encode.return_value = [
        [1],
        [1, 2],
        [1, 2, 3],
        [1, 2, 3, 4],
        [1, 2, 3, 4, 5],
    ]
    schema = types.Schema(
        type=types.Type.OBJECT,
        format='schema_format',
        description='Recipe schema',
        enum=['schema_enum1', 'schema_enum2'],
        properties={
            'recipe_name': types.Schema(
                type=types.Type.STRING,
                description='Name of the recipe',
            )
        },
        items=types.Schema(
            type=types.Type.STRING,
            description='Item in the recipe',
        ),
        example={
            'recipe_example': types.Schema(
                type=types.Type.STRING,
                description='example in the recipe',
            )
        },
        required=['recipe_name'],
    )
    config = types.CountTokensConfig(
        generation_config=types.GenerationConfig(response_schema=schema)
    )
    result = self.tokenizer.count_tokens(
        'Generate a recipe for chocolate chip cookies.', config=config
    )
    self.assertEqual(result.total_tokens, 15)
    self.mock_tokenizer.encode.assert_called_once_with([
        'Generate a recipe for chocolate chip cookies.',
        'schema_format',
        'Recipe schema',
        'schema_enum1',
        'schema_enum2',
        'recipe_name',
        'Item in the recipe',
        'recipe_name',
        'Name of the recipe',
        'recipe_example',
    ])

  def test_count_tokens_with_unsupported_fields_logs_warning(self):
    self.mock_tokenizer.encode.return_value = [[1, 2, 3]]
    content_with_unsupported = types.Content(
        role='user',
        parts=[
            types.Part(text='hello'),
            # executable_code is not supported by _TextsAccumulator
            types.Part(
                executable_code=types.ExecutableCode(
                    language='PYTHON', code='print(1)'
                )
            ),
        ],
    )
    with self.assertLogs('google_genai.local_tokenizer', level='WARNING') as cm:
      self.tokenizer.count_tokens(content_with_unsupported)
      self.assertIn(
          'Content contains unsupported types for token counting', cm.output[0]
      )

  def test_compute_tokens_simple_string(self):
    mock_spt = MagicMock()
    mock_spt.pieces = [
        MagicMock(id=1, piece='He'),
        MagicMock(id=2, piece='llo'),
        MagicMock(id=3, piece=' world'),
    ]
    self.mock_tokenizer.EncodeAsImmutableProto.return_value = [mock_spt]
    result = self.tokenizer.compute_tokens('Hello world')
    self.assertEqual(len(result.tokens_info), 1)
    self.assertEqual(result.tokens_info[0].token_ids, [1, 2, 3])
    self.assertEqual(result.tokens_info[0].tokens, [b'He', b'llo', b' world'])
    self.assertEqual(result.tokens_info[0].role, 'user')
    self.mock_tokenizer.EncodeAsImmutableProto.assert_called_once_with(
        ['Hello world']
    )

  def test_compute_tokens_with_chat_history(self):
    mock_spt1 = MagicMock()
    mock_spt1.pieces = [MagicMock(id=1, piece='Hello')]
    mock_spt2 = MagicMock()
    mock_spt2.pieces = [
        MagicMock(id=2, piece='Hi'),
        MagicMock(id=3, piece=' there!'),
    ]
    self.mock_tokenizer.EncodeAsImmutableProto.return_value = [
        mock_spt1,
        mock_spt2,
    ]
    history = [
        types.Content(role='user', parts=[types.Part(text='Hello')]),
        types.Content(role='model', parts=[types.Part(text='Hi there!')]),
    ]
    result = self.tokenizer.compute_tokens(history)
    self.assertEqual(len(result.tokens_info), 2)
    self.assertEqual(result.tokens_info[0].token_ids, [1])
    self.assertEqual(result.tokens_info[0].tokens, [b'Hello'])
    self.assertEqual(result.tokens_info[0].role, 'user')
    self.assertEqual(result.tokens_info[1].token_ids, [2, 3])
    self.assertEqual(result.tokens_info[1].tokens, [b'Hi', b' there!'])
    self.assertEqual(result.tokens_info[1].role, 'model')
    self.mock_tokenizer.EncodeAsImmutableProto.assert_called_once_with(
        ['Hello', 'Hi there!']
    )

  def test_compute_tokens_with_byte_tokens(self):
    mock_spt = MagicMock()
    mock_spt.pieces = [
        MagicMock(id=1, piece='<0x48>'),
        MagicMock(id=2, piece='ello'),
    ]
    self.mock_tokenizer.EncodeAsImmutableProto.return_value = [mock_spt]
    self.tokenizer._model_proto = sentencepiece_model_pb2.ModelProto(
        pieces=[
            sentencepiece_model_pb2.ModelProto.SentencePiece(),
            sentencepiece_model_pb2.ModelProto.SentencePiece(
                type=sentencepiece_model_pb2.ModelProto.SentencePiece.Type.BYTE
            ),
            sentencepiece_model_pb2.ModelProto.SentencePiece(
                type=sentencepiece_model_pb2.ModelProto.SentencePiece.Type.NORMAL
            ),
        ]
    )
    result = self.tokenizer.compute_tokens('Hello')
    self.assertEqual(len(result.tokens_info), 1)
    self.assertEqual(result.tokens_info[0].token_ids, [1, 2])
    self.assertEqual(result.tokens_info[0].tokens, [b'H', b'ello'])
    self.mock_tokenizer.EncodeAsImmutableProto.assert_called_once_with(
        ['Hello']
    )


class TestParseHexByte(unittest.TestCase):

  def test_valid_hex(self):
    self.assertEqual(local_tokenizer._parse_hex_byte('<0x41>'), 65)
    self.assertEqual(local_tokenizer._parse_hex_byte('<0xFF>'), 255)
    self.assertEqual(local_tokenizer._parse_hex_byte('<0x00>'), 0)

  def test_invalid_length(self):
    with self.assertRaisesRegex(ValueError, 'Invalid byte length'):
      local_tokenizer._parse_hex_byte('<0x41')
    with self.assertRaisesRegex(ValueError, 'Invalid byte length'):
      local_tokenizer._parse_hex_byte('<0x411>')

  def test_invalid_format(self):
    with self.assertRaisesRegex(ValueError, 'Invalid byte format'):
      local_tokenizer._parse_hex_byte(' 0x41>')
    with self.assertRaisesRegex(ValueError, 'Invalid byte format'):
      local_tokenizer._parse_hex_byte('<0x41 ')

  def test_invalid_hex_value(self):
    with self.assertRaisesRegex(ValueError, 'Invalid hex value'):
      local_tokenizer._parse_hex_byte('<0xFG>')



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Union, Optional
from typing_extensions import Literal, Annotated, TypeAlias

from .._utils import PropertyInfo
from .._models import BaseModel
from .text_content import TextContent
from .image_content import ImageContent

__all__ = ["ThoughtContent", "Summary"]

Summary: TypeAlias = Annotated[Union[TextContent, ImageContent], PropertyInfo(discriminator="type")]


class ThoughtContent(BaseModel):
    """A thought content block."""

    type: Literal["thought"]

    signature: Optional[str] = None
    """Signature to match the backend source to be part of the generation."""

    summary: Optional[List[Summary]] = None
    """A summary of the thought."""



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Union, Optional
from typing_extensions import Literal, TypeAlias

from .._models import BaseModel
from .text_content import TextContent
from .image_content import ImageContent

__all__ = ["MCPServerToolResultContent", "Result", "ResultItems", "ResultItemsItem"]

ResultItemsItem: TypeAlias = Union[TextContent, ImageContent]


class ResultItems(BaseModel):
    items: Optional[List[ResultItemsItem]] = None


Result: TypeAlias = Union[ResultItems, str, object]


class MCPServerToolResultContent(BaseModel):
    """MCPServer tool result content."""

    call_id: str
    """ID to match the ID from the MCP server tool call block."""

    result: Result
    """The result of the tool call."""

    type: Literal["mcp_server_tool_result"]

    name: Optional[str] = None
    """Name of the tool which is called for this specific tool call."""

    server_name: Optional[str] = None
    """The name of the used MCP server."""



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import pytest

IS_NOT_GITHUB_ACTIONS = os.getenv('GITHUB_ACTIONS') != 'true'


@pytest.mark.skipif(IS_NOT_GITHUB_ACTIONS,
                    reason='This test is only run on GitHub Actions.')
def test_library_can_be_imported_without_optional_dependencies():
  """Tests that the library can be imported without optional dependencies.
  """
  from google import genai
  from google.genai import types



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""[Experimental] Text Only Local Tokenizer."""

import logging
from typing import Any, Iterable
from typing import Optional, Union

from sentencepiece import sentencepiece_model_pb2

from . import _common
from . import _local_tokenizer_loader as loader
from . import _transformers as t
from . import types

logger = logging.getLogger("google_genai.local_tokenizer")

__all__ = [
    "_parse_hex_byte",
    "_token_str_to_bytes",
    "LocalTokenizer",
    "_TextsAccumulator",
]


class _TextsAccumulator:
  """Accumulates countable texts from `Content` and `Tool` objects.

  This class is responsible for traversing complex `Content` and `Tool`
  objects and extracting all the text content that should be included when
  calculating token counts.

  A key feature of this class is its ability to detect unsupported fields in
  `Content` objects. If a user provides a `Content` object with fields that
  this local tokenizer doesn't recognize (e.g., new fields added in a future
  API update), this class will log a warning.

  The detection mechanism for `Content` objects works by recursively building
  a "counted" version of the input object. This "counted" object only
  contains the data that was successfully processed and added to the text
  list for tokenization. After traversing the input, the original `Content`
  object is compared to the "counted" object. If they don't match, it
  signifies the presence of unsupported fields, and a warning is logged.
  """

  def __init__(self) -> None:
    self._texts: list[str] = []

  def get_texts(self) -> Iterable[str]:
    return self._texts

  def add_contents(self, contents: Iterable[types.Content]) -> None:
    for content in contents:
      self.add_content(content)

  def add_content(self, content: types.Content) -> None:
    counted_content = types.Content(parts=[], role=content.role)
    if content.parts:
      for part in content.parts:
        assert counted_content.parts is not None
        counted_part = types.Part()
        if part.file_data is not None or part.inline_data is not None:
          raise ValueError(
              "LocalTokenizers do not support non-text content types."
          )
        if part.video_metadata is not None:
          counted_part.video_metadata = part.video_metadata
        if part.function_call is not None:
          self.add_function_call(part.function_call)
          counted_part.function_call = part.function_call
        if part.function_response is not None:
          self.add_function_response(part.function_response)
          counted_part.function_response = part.function_response
        if part.text is not None:
          counted_part.text = part.text
          self._texts.append(part.text)
        counted_content.parts.append(counted_part)

    if content.model_dump(exclude_none=True) != counted_content.model_dump(
        exclude_none=True
    ):
      logger.warning(
          "Content contains unsupported types for token counting. Supported"
          f" fields {counted_content}. Got {content}."
      )

  def add_function_call(self, function_call: types.FunctionCall) -> None:
    """Processes a function call and adds relevant text to the accumulator.

    Args:
        function_call: The function call to process.
    """
    if function_call.name:
      self._texts.append(function_call.name)
    counted_function_call = types.FunctionCall(name=function_call.name)
    if function_call.args:
      counted_args = self._dict_traverse(function_call.args)
      counted_function_call.args = counted_args

  def add_tool(self, tool: types.Tool) -> types.Tool:
    counted_tool = types.Tool(function_declarations=[])
    if tool.function_declarations:
      for function_declaration in tool.function_declarations:
        counted_function_declaration = self._function_declaration_traverse(
            function_declaration
        )
        if counted_tool.function_declarations is None:
          counted_tool.function_declarations = []
        counted_tool.function_declarations.append(counted_function_declaration)

    return counted_tool

  def add_tools(self, tools: Iterable[types.Tool]) -> None:
    for tool in tools:
      self.add_tool(tool)

  def add_function_responses(
      self, function_responses: Iterable[types.FunctionResponse]
  ) -> None:
    for function_response in function_responses:
      self.add_function_response(function_response)

  def add_function_response(
      self, function_response: types.FunctionResponse
  ) -> None:
    counted_function_response = types.FunctionResponse()
    if function_response.name:
      self._texts.append(function_response.name)
      counted_function_response.name = function_response.name
    if function_response.response:
      counted_response = self._dict_traverse(function_response.response)
      counted_function_response.response = counted_response

  def _function_declaration_traverse(
      self, function_declaration: types.FunctionDeclaration
  ) -> types.FunctionDeclaration:
    counted_function_declaration = types.FunctionDeclaration()
    if function_declaration.name:
      self._texts.append(function_declaration.name)
      counted_function_declaration.name = function_declaration.name
    if function_declaration.description:
      self._texts.append(function_declaration.description)
      counted_function_declaration.description = (
          function_declaration.description
      )
    if function_declaration.parameters:
      counted_parameters = self.add_schema(function_declaration.parameters)
      counted_function_declaration.parameters = counted_parameters
    if function_declaration.response:
      counted_response = self.add_schema(function_declaration.response)
      counted_function_declaration.response = counted_response
    return counted_function_declaration

  def add_schema(self, schema: types.Schema) -> types.Schema:
    """Processes a schema and adds relevant text to the accumulator.

    Args:
        schema: The schema to process.

    Returns:
        The new schema object with only countable fields.
    """
    counted_schema = types.Schema()
    if schema.type:
      counted_schema.type = schema.type
    if schema.title:
      counted_schema.title = schema.title
    if schema.default is not None:
      counted_schema.default = schema.default
    if schema.format:
      self._texts.append(schema.format)
      counted_schema.format = schema.format
    if schema.description:
      self._texts.append(schema.description)
      counted_schema.description = schema.description
    if schema.enum:
      self._texts.extend(schema.enum)
      counted_schema.enum = schema.enum
    if schema.required:
      self._texts.extend(schema.required)
      counted_schema.required = schema.required
    if schema.property_ordering:
      counted_schema.property_ordering = schema.property_ordering
    if schema.items:
      counted_schema_items = self.add_schema(schema.items)
      counted_schema.items = counted_schema_items
    if schema.properties:
      d = {}
      for key, value in schema.properties.items():
        self._texts.append(key)
        counted_value = self.add_schema(value)
        d[key] = counted_value
      counted_schema.properties = d
    if schema.example:
      counted_schema_example = self._any_traverse(schema.example)
      counted_schema.example = counted_schema_example
    return counted_schema

  def _dict_traverse(self, d: dict[str, Any]) -> dict[str, Any]:
    """Processes a dict and adds relevant text to the accumulator.

    Args:
        d: The dict to process.

    Returns:
        The new dict object with only countable fields.
    """
    counted_dict = {}
    self._texts.extend(list(d.keys()))
    for key, val in d.items():
      counted_dict[key] = self._any_traverse(val)
    return counted_dict

  def _any_traverse(self, value: Any) -> Any:
    """Processes a value and adds relevant text to the accumulator.

    Args:
        value: The value to process.

    Returns:
        The new value with only countable fields.
    """
    if isinstance(value, str):
      self._texts.append(value)
      return value
    elif isinstance(value, dict):
      return self._dict_traverse(value)
    elif isinstance(value, list):
      return [self._any_traverse(item) for item in value]
    else:
      return value


def _token_str_to_bytes(
    token: str, type: sentencepiece_model_pb2.ModelProto.SentencePiece.Type
) -> bytes:
  if type == sentencepiece_model_pb2.ModelProto.SentencePiece.Type.BYTE:
    return _parse_hex_byte(token).to_bytes(length=1, byteorder="big")
  else:
    return token.replace("▁", " ").encode("utf-8")


def _parse_hex_byte(token: str) -> int:
  """Parses a hex byte string of the form '<0xXX>' and returns the integer value.

  Raises ValueError if the input is malformed or the byte value is invalid.
  """

  if len(token) != 6:
    raise ValueError(f"Invalid byte length: {token}")
  if not token.startswith("<0x") or not token.endswith(">"):
    raise ValueError(f"Invalid byte format: {token}")

  try:
    val = int(token[3:5], 16)  # Parse the hex part directly
  except ValueError:
    raise ValueError(f"Invalid hex value: {token}")

  if val >= 256:
    raise ValueError(f"Byte value out of range: {token}")

  return val


class LocalTokenizer:
  """[Experimental] Text Only Local Tokenizer.

  This class provides a local tokenizer for text only token counting.

  LIMITATIONS:
  - Only supports text based tokenization and no multimodal tokenization.
  - Forward compatibility depends on the open-source tokenizer models for future
  Gemini versions.
  - For token counting of tools and response schemas, the `LocalTokenizer` only
  supports `types.Tool` and `types.Schema` objects. Python functions or Pydantic
  models cannot be passed directly.
  """

  def __init__(self, model_name: str):
    self._tokenizer_name = loader.get_tokenizer_name(model_name)
    self._model_proto = loader.load_model_proto(self._tokenizer_name)
    self._tokenizer = loader.get_sentencepiece(self._tokenizer_name)

  @_common.experimental_warning(
      "The SDK's local tokenizer implementation is experimental and may change"
      " in the future. It only supports text based tokenization."
  )
  def count_tokens(
      self,
      contents: Union[types.ContentListUnion, types.ContentListUnionDict],
      *,
      config: Optional[types.CountTokensConfigOrDict] = None,
  ) -> types.CountTokensResult:
    """Counts the number of tokens in a given text.

    Args:
      contents: The contents to tokenize.
      config: The configuration for counting tokens.

    Returns:
      A `CountTokensResult` containing the total number of tokens.

    Usage:

    .. code-block:: python

      from google import genai
      tokenizer = genai.LocalTokenizer(model_name='gemini-2.0-flash-001')
      result = tokenizer.count_tokens("What is your name?")
      print(result)
      # total_tokens=5
    """
    processed_contents = t.t_contents(contents)
    text_accumulator = _TextsAccumulator()
    config = types.CountTokensConfig.model_validate(config or {})
    text_accumulator.add_contents(processed_contents)
    if config.tools:
      text_accumulator.add_tools(config.tools)
    if config.generation_config and config.generation_config.response_schema:
      text_accumulator.add_schema(config.generation_config.response_schema)
    if config.system_instruction:
      text_accumulator.add_contents(t.t_contents([config.system_instruction]))
    tokens_list = self._tokenizer.encode(list(text_accumulator.get_texts()))
    return types.CountTokensResult(
        total_tokens=sum(len(tokens) for tokens in tokens_list)
    )

  @_common.experimental_warning(
      "The SDK's local tokenizer implementation is experimental and may change"
      " in the future. It only supports text based tokenization."
  )
  def compute_tokens(
      self,
      contents: Union[types.ContentListUnion, types.ContentListUnionDict],
  ) -> types.ComputeTokensResult:
    """Computes the tokens ids and string pieces in the input.

    Args:
      contents: The contents to tokenize.

    Returns:
      A `ComputeTokensResult` containing the token information.

    Usage:

    .. code-block:: python

      from google import genai
      tokenizer = genai.LocalTokenizer(model_name='gemini-2.0-flash-001')
      result = tokenizer.compute_tokens("What is your name?")
      print(result)
      # tokens_info=[TokensInfo(token_ids=[279, 329, 1313, 2508, 13], tokens=[b' What', b' is', b' your', b' name', b'?'], role='user')]
    """
    processed_contents = t.t_contents(contents)
    text_accumulator = _TextsAccumulator()
    for content in processed_contents:
      text_accumulator.add_content(content)
    tokens_protos = self._tokenizer.EncodeAsImmutableProto(
        text_accumulator.get_texts()
    )

    roles = []
    for content in processed_contents:
      if content.parts:
        for _ in content.parts:
          roles.append(content.role)

    token_infos = []
    for tokens_proto, role in zip(tokens_protos, roles):
      token_infos.append(
          types.TokensInfo(
              token_ids=[piece.id for piece in tokens_proto.pieces],
              tokens=[
                  _token_str_to_bytes(
                      piece.piece, self._model_proto.pieces[piece.id].type
                  )
                  for piece in tokens_proto.pieces
              ],
              role=role,
          )
      )
    return types.ComputeTokensResult(tokens_info=token_infos)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import pathlib

from pydantic import BaseModel, ValidationError, Field, ConfigDict
from typing import Literal, List, Optional, Union, Set
from datetime import datetime
import pytest
import json
import logging
import sys
from ... import _transformers as t
from ... import errors
from ... import types
from .. import pytest_helper
from enum import Enum

GEMINI_FLASH_LATEST = 'gemini-2.5-flash'
GEMINI_FLASH_2_0 = 'gemini-2.0-flash-001'
GEMINI_FLASH_IMAGE_LATEST = 'gemini-2.5-flash-image'

IMAGE_PNG_FILE_PATH = pathlib.Path(__file__).parent / '../data/google.png'
image_bytes = IMAGE_PNG_FILE_PATH.read_bytes()

AUDIO_WAV_FILE_PATH = pathlib.Path(__file__).parent / '../data/voice_sample.wav'
audio_bytes = AUDIO_WAV_FILE_PATH.read_bytes()


safety_settings_with_method = [
    {
        'category': 'HARM_CATEGORY_HATE_SPEECH',
        'threshold': 'BLOCK_ONLY_HIGH',
        'method': 'SEVERITY',
    },
    {
        'category': 'HARM_CATEGORY_DANGEROUS_CONTENT',
        'threshold': 'BLOCK_LOW_AND_ABOVE',
        'method': 'PROBABILITY',
    },
]

test_http_options = {'api_version': 'v1', 'headers': {'test': 'headers'}}


class InstrumentEnum(Enum):
  PERCUSSION = 'Percussion'
  STRING = 'String'
  WOODWIND = 'Woodwind'
  BRASS = 'Brass'
  KEYBOARD = 'Keyboard'


test_table: list[pytest_helper.TestTableItem] = [
    pytest_helper.TestTableItem(
        name='test_http_options_in_method',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('What is your name?'),
            config={
                'http_options': test_http_options,
            },
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_union_contents_is_string',
        override_replay_id='test_sync',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST, contents='Tell me a story in 300 words.'
        ),
        has_union=True,
    ),
    pytest_helper.TestTableItem(
        name='test_union_contents_is_content',
        override_replay_id='test_sync',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=types.Content(
                role='user',
                parts=[types.Part(text='Tell me a story in 300 words.')],
            ),
        ),
        has_union=True,
    ),
    pytest_helper.TestTableItem(
        name='test_union_contents_is_parts',
        override_replay_id='test_sync',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=[types.Part(text='Tell me a story in 300 words.')],
        ),
        has_union=True,
    ),
    pytest_helper.TestTableItem(
        name='test_union_contents_is_part',
        override_replay_id='test_sync',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=types.Part(text='Tell me a story in 300 words.'),
        ),
        has_union=True,
    ),
    pytest_helper.TestTableItem(
        name='test_sync_content_list',
        override_replay_id='test_sync',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=[
                types.Content(
                    role='user',
                    parts=[types.Part(text='Tell me a story in 300 words.')],
                )
            ],
        ),
    ),
    # You need to enable llama API in Vertex AI Model Garden.
    pytest_helper.TestTableItem(
        name='test_llama',
        parameters=types._GenerateContentParameters(
            model='meta/llama-3.2-90b-vision-instruct-maas',
            contents=t.t_contents('What is your name?'),
        ),
        exception_if_mldev='404',
        skip_in_api_mode='it will encounter 403 for api mode',
    ),
    pytest_helper.TestTableItem(
        name='test_system_instructions',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('high'),
            config={
                'system_instruction': t.t_content(
                    'I say high, you say low'
                )
            },
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_labels',
        exception_if_mldev='not supported',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('What is your name?'),
            config={
                'labels': {'label1': 'value1', 'label2': 'value2'},
            },
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_simple_shared_generation_config',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('What is your name?'),
            config={
                'max_output_tokens': 100,
                'top_k': 2,
                'temperature': 0.5,
                'top_p': 0.5,
                'response_mime_type': 'application/json',
                'stop_sequences': ['\n'],
                'candidate_count': 2,
                'seed': 42,
            },
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_2_candidates_gemini_2_5_flash',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('Tell me a story in 30 words.'),
            config={
                'candidate_count': 2,
            },
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_safety_settings_on_difference',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('What is your name?'),
            config={
                'safety_settings': safety_settings_with_method,
            },
        ),
        exception_if_mldev='method',
    ),
    pytest_helper.TestTableItem(
        name='test_penalty',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_2_0,
            contents=t.t_contents('Tell me a story in 30 words.'),
            config={
                'presence_penalty': 0.5,
                'frequency_penalty': 0.5,
            },
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_penalty_gemini_2_0_flash',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_2_0,
            contents=t.t_contents('Tell me a story in 30 words.'),
            config={
                'presence_penalty': 0.5,
                'frequency_penalty': 0.5,
            },
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_google_search_tool',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('Why is the sky blue?'),
            config=types.GenerateContentConfig(
                tools=[types.Tool(google_search=types.GoogleSearch())]
            ),
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_google_maps_tool',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('Find restaurants near me.'),
            config=types.GenerateContentConfig(
                tools=[{'google_maps': {}}],
                tool_config={
                    "retrieval_config": {
                        "lat_lng": {
                            "latitude": 37.421993,
                            "longitude": -122.079725,
                        }
                    }
                }
            ),
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_google_search_tool_with_time_range_filter',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('What is the QQQ stock price?'),
            config=types.GenerateContentConfig(
                tools=[
                    types.Tool(
                        google_search=types.GoogleSearch(
                            time_range_filter=types.Interval(
                                start_time=datetime.fromisoformat(
                                    '2025-05-01T00:00:00Z'
                                ),
                                end_time=datetime.fromisoformat(
                                    '2025-05-03T00:00:00Z'
                                ),
                            )
                        )
                    )
                ]
            ),
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_google_search_tool_with_exclude_domains',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('Why is the sky blue?'),
            config=types.GenerateContentConfig(
                tools=[
                    types.Tool(
                        google_search=types.GoogleSearch(
                            exclude_domains=['amazon.com', 'facebook.com']
                        )
                    )
                ]
            ),
        ),
        exception_if_mldev='not supported in',
    ),
    pytest_helper.TestTableItem(
        name='test_google_search_tool_with_blocking_confidence',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('Why is the sky blue?'),
            config=types.GenerateContentConfig(
                tools=[
                    types.Tool(
                        google_search=types.GoogleSearch(
                            blocking_confidence=types.PhishBlockThreshold.BLOCK_LOW_AND_ABOVE,
                        )
                    )
                ]
            ),
        ),
        exception_if_mldev='not supported in',
    ),
    pytest_helper.TestTableItem(
        name='test_enterprise_web_search_tool',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('Why is the sky blue?'),
            config=types.GenerateContentConfig(
                tools=[
                    types.Tool(
                        enterprise_web_search=types.EnterpriseWebSearch()
                    )
                ]
            ),
        ),
        exception_if_mldev='not supported in',
    ),
    pytest_helper.TestTableItem(
        name='test_enterprise_web_search_tool_with_exclude_domains',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('Why is the sky blue?'),
            config=types.GenerateContentConfig(
                tools=[
                    types.Tool(
                        enterprise_web_search=types.EnterpriseWebSearch(
                            exclude_domains=['amazon.com', 'facebook.com']
                        )
                    )
                ]
            ),
        ),
        exception_if_mldev='not supported in',
    ),
    pytest_helper.TestTableItem(
        name='test_enterprise_web_search_tool_with_blocking_confidence',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('Why is the sky blue?'),
            config=types.GenerateContentConfig(
                tools=[
                    types.Tool(
                        enterprise_web_search=types.EnterpriseWebSearch(
                            blocking_confidence=types.PhishBlockThreshold.BLOCK_LOW_AND_ABOVE,
                        )
                    )
                ]
            ),
        ),
        exception_if_mldev='not supported in',
    ),
    pytest_helper.TestTableItem(
        name='test_speech_with_config',
        parameters=types._GenerateContentParameters(
            model='gemini-2.5-flash-preview-tts',
            contents=t.t_contents(
                'Produce a speech response saying "Cheese"'
            ),
            config=types.GenerateContentConfig(
                response_modalities=['audio'],
                speech_config=types.SpeechConfig(
                    voice_config=types.VoiceConfig(
                        prebuilt_voice_config=types.PrebuiltVoiceConfig(
                            voice_name='charon'
                        )
                    )
                ),
            ),
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_speech_with_multi_speaker_voice_config',
        parameters=types._GenerateContentParameters(
            model='gemini-2.5-flash-preview-tts',
            contents=t.t_contents(
                'Alice says "Hi", Bob replies with "what\'s up"?'
            ),
            config=types.GenerateContentConfig(
                response_modalities=['audio'],
                speech_config=types.SpeechConfig(
                    multi_speaker_voice_config=types.MultiSpeakerVoiceConfig(
                        speaker_voice_configs=[
                            types.SpeakerVoiceConfig(
                                speaker='Alice',
                                voice_config=types.VoiceConfig(
                                    prebuilt_voice_config=types.PrebuiltVoiceConfig(
                                        voice_name='leda'
                                    )
                                ),
                            ),
                            types.SpeakerVoiceConfig(
                                speaker='Bob',
                                voice_config=types.VoiceConfig(
                                    prebuilt_voice_config=types.PrebuiltVoiceConfig(
                                        voice_name='kore'
                                    )
                                ),
                            ),
                        ],
                    )
                ),
            ),
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_speech_error_with_speech_config_and_multi_speech_config',
        exception_if_vertex='mutually exclusive',
        exception_if_mldev='mutually exclusive',
        parameters=types._GenerateContentParameters(
            model='gemini-2.5-flash-preview-tts',
            contents=t.t_contents(
                'Alice says "Hi", Bob replies with "what\'s up"?'
            ),
            config=types.GenerateContentConfig(
                response_modalities=['audio'],
                speech_config=types.SpeechConfig(
                    voice_config=types.VoiceConfig(
                        prebuilt_voice_config=types.PrebuiltVoiceConfig(
                            voice_name='puck'
                        )
                    ),
                    multi_speaker_voice_config=types.MultiSpeakerVoiceConfig(
                        speaker_voice_configs=[
                            types.SpeakerVoiceConfig(
                                speaker='Alice',
                                voice_config=types.VoiceConfig(
                                    prebuilt_voice_config=types.PrebuiltVoiceConfig(
                                        voice_name='leda'
                                    )
                                ),
                            ),
                            types.SpeakerVoiceConfig(
                                speaker='Bob',
                                voice_config=types.VoiceConfig(
                                    prebuilt_voice_config=types.PrebuiltVoiceConfig(
                                        voice_name='kore'
                                    )
                                ),
                            ),
                        ],
                    ),
                ),
            ),
        ),
    ),
    pytest_helper.TestTableItem(
        name='test_union_speech_string_config',
        parameters=types._GenerateContentParameters(
            model='gemini-2.5-flash-preview-tts',
            contents='Say hello!',
            config=types.GenerateContentConfig(
                response_modalities=['audio'], speech_config='charon'
            ),
        ),
        has_union=True,
    ),
    pytest_helper.TestTableItem(
        name='test_audio_timestamp',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=[
                types.Content(
                    role='user',
                    parts=[
                        types.Part(
                            file_data=types.FileData(
                                file_uri='gs://cloud-samples-data/generative-ai/audio/pixel.mp3',
                                mime_type='audio/mpeg',
                            )
                        ),
                        types.Part(
                            text="""Can you transcribe this interview, in the
                           format of timecode, speaker, caption. Use speaker A, 
                           speaker B, etc. to identify speakers."""
                        ),
                    ],
                )
            ],
            config=types.GenerateContentConfig(audio_timestamp=True),
        ),
        exception_if_mldev='not supported',
    ),
    pytest_helper.TestTableItem(
        name='test_response_schema_with_default',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('What is your name?'),
            config={
                'response_mime_type': 'application/json',
                'response_schema': {
                    'type': 'OBJECT',
                    'properties': {
                        'name': {
                            'type': 'STRING',
                            'default': 'default_name',
                        },
                    },
                },
            },
        ),
        ignore_keys=['parsed'],
    ),
    pytest_helper.TestTableItem(
        name='test_invalid_model_parameter_path',
        parameters=types._GenerateContentParameters(
            model='gemini-2.5-flash/../../upload/v1beta/files',
            contents=t.t_contents('What is your name?'),
        ),
        exception_if_vertex='invalid model parameter',
        exception_if_mldev='invalid model parameter',
    ),
    pytest_helper.TestTableItem(
        name='test_invalid_model_parameter_question_mark',
        parameters=types._GenerateContentParameters(
            model='gemini-2.5-flash?',
            contents=t.t_contents('What is your name?'),
        ),
        exception_if_vertex='invalid model parameter',
        exception_if_mldev='invalid model parameter',
    ),
    pytest_helper.TestTableItem(
        name='test_invalid_model_parameter_ampersand',
        parameters=types._GenerateContentParameters(
            model='gemini-2.5-flash&',
            contents=t.t_contents('What is your name?'),
        ),
        exception_if_vertex='invalid model parameter',
        exception_if_mldev='invalid model parameter',
    ),
    pytest_helper.TestTableItem(
        name='test_enhanced_civic_answers',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents(
                'Summarize the evidence that confirms widespread voter fraud was'
                ' the reason the last national election results were'
                ' inaccurate.'
            ),
            config={
                'enable_enhanced_civic_answers': True,
            },
        ),
        exception_if_vertex='not supported',
    ),
    pytest_helper.TestTableItem(
        name='test_model_armor_config',
        parameters=types._GenerateContentParameters(
            model=GEMINI_FLASH_LATEST,
            contents=t.t_contents('What is your name?'),
            config={
                'model_armor_config': {
                    'prompt_template_name': '',
                    'response_template_name': '',
                    # Intentionally left blank just to test that the SDK doesn't
                    # throw an exception.
                },
            },
        ),
        exception_if_mldev='not supported',
    ),
]

pytestmark = pytest_helper.setup(
    file=__file__,
    globals_for_file=globals(),
    test_method='models.generate_content',
    test_table=test_table,
)
pytest_plugins = ('pytest_asyncio',)


def test_sync_with_headers(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Tell me a story in 300 words.',
  )
  assert response.sdk_http_response.headers is not None
  assert response.sdk_http_response.body is None


def test_sync_with_full_response(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Tell me a story in 300 words.',
      config={
          'should_return_http_response': True,
      },
  )
  print(response.sdk_http_response.body)
  assert response.sdk_http_response.headers is not None
  assert response.sdk_http_response.body is not None
  assert 'candidates' in response.sdk_http_response.body
  assert 'content' in response.sdk_http_response.body
  assert 'parts' in response.sdk_http_response.body
  assert 'usageMetadata' in response.sdk_http_response.body

@pytest.mark.asyncio
async def test_async(client):
  response = await client.aio.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Tell me a story in 300 words.',
      config={
          'http_options': test_http_options,
      },
  )
  assert response.text


@pytest.mark.asyncio
async def test_async_with_headers(client):
  response = await client.aio.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Tell me a story in 300 words.',
  )
  assert response.sdk_http_response.headers is not None
  assert response.sdk_http_response.body is None


@pytest.mark.asyncio
async def test_async_with_full_response(client):
  response = await client.aio.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Tell me a story in 300 words.',
      config={
          'should_return_http_response': True,
      },
  )
  assert response.sdk_http_response.headers is not None
  assert response.sdk_http_response.body is not None
  assert 'candidates' in response.sdk_http_response.body
  assert 'content' in response.sdk_http_response.body
  assert 'parts' in response.sdk_http_response.body
  assert 'usageMetadata' in response.sdk_http_response.body


def test_sync_stream(client):
  response = client.models.generate_content_stream(
      model=GEMINI_FLASH_LATEST,
      contents='Tell me a story in 300 words.',
      config={
          'http_options': test_http_options,
      },
  )
  chunks = 0
  for part in response:
    chunks += 1
    assert part.text is not None or part.candidates[0].finish_reason

  assert chunks >= 1


def test_sync_stream_with_should_return_http_headers(client):
  response = client.models.generate_content_stream(
      model=GEMINI_FLASH_LATEST,
      contents='Tell me a story in 300 words.',
      config={
          'http_options': test_http_options,
      },
  )
  chunks = 0
  for part in response:
    chunks += 1
    assert part.text is not None or part.candidates[0].finish_reason
    assert part.sdk_http_response.headers is not None
  assert chunks >= 1


def test_sync_stream_with_non_text_modality(client):
  response = client.models.generate_content_stream(
      model='gemini-2.0-flash-preview-image-generation',
      contents=(
          'Generate an image of the Eiffel tower with fireworks in the'
          ' background.'
      ),
      config={
          'response_modalities': ['IMAGE', 'TEXT'],
      },
  )
  chunks = 0
  for chunk in response:
    chunks += 1
    if chunk.candidates[0].finish_reason is not None:
      continue
    for part in chunk.parts:
      assert part.text is not None or part.inline_data is not None

  assert chunks >= 1


@pytest.mark.asyncio
async def test_async_stream(client):
  chunks = 0
  async for part in await client.aio.models.generate_content_stream(
      model=GEMINI_FLASH_LATEST, contents='Tell me a story in 300 words.',
      config={
          'http_options': test_http_options,
      },
  ):
    chunks += 1
    assert part.text is not None or part.candidates[0].finish_reason

  assert chunks >= 1


@pytest.mark.asyncio
async def test_async_stream_with_headers(client):
  chunks = 0
  async for part in await client.aio.models.generate_content_stream(
      model=GEMINI_FLASH_LATEST, contents='Tell me a story in 300 words.',
      config={
          'http_options': test_http_options,
      },
  ):
    chunks += 1
    assert part.text is not None or part.candidates[0].finish_reason
    assert part.sdk_http_response.headers is not None

  assert chunks >= 1


@pytest.mark.asyncio
async def test_async_stream_with_non_text_modality(client):
  chunks = 0
  async for chunk in await client.aio.models.generate_content_stream(
      model=GEMINI_FLASH_IMAGE_LATEST,
      contents=(
          'Generate an image of the Eiffel tower with fireworks in the'
          ' background.'
      ),
      config={
          'response_modalities': ['IMAGE', 'TEXT'],
      },
  ):
    chunks += 1
    if chunk.candidates[0].finish_reason is not None:
      continue
    for part in chunk.parts:
      assert part.text is not None or part.inline_data is not None

  assert chunks >= 1


def test_simple_shared_generation_config_stream(client):
  chunks = 0
  for chunk in client.models.generate_content_stream(
      model=GEMINI_FLASH_LATEST,
      contents='tell me a story in 300 words',
      config={
          'max_output_tokens': 1000,
          'top_k': 2,
          'temperature': 0.5,
          'top_p': 0.5,
          'response_mime_type': 'application/json',
          'stop_sequences': ['\n'],
          'seed': 42,
      },
  ):
    chunks += 1
    assert (
        chunk.text is not None or chunk.candidates[0].finish_reason
    ), f'vertexai: {client._api_client.vertexai}, {chunk.candidate[0]}'
  assert chunks >= 1


@pytest.mark.asyncio
async def test_simple_shared_generation_config_async(client):
  response = await client.aio.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='tell me a story in 300 words',
      config={
          'max_output_tokens': 4000,
          'top_k': 2,
          'temperature': 0.5,
          'top_p': 0.5,
          'response_mime_type': 'application/json',
          'stop_sequences': ['\n'],
          'seed': 42,
      },
  )


@pytest.mark.asyncio
async def test_simple_shared_generation_config_stream_async(client):
  chunks = 0
  async for part in await client.aio.models.generate_content_stream(
      model=GEMINI_FLASH_2_0,
      contents='tell me a story in 300 words',
      config={
          'max_output_tokens': 400,
          'top_k': 2,
          'temperature': 0.5,
          'top_p': 0.5,
          'response_mime_type': 'application/json',
          'stop_sequences': ['\n'],
          'seed': 42,
      },
  ):
    chunks += 1
    assert part.text is not None or part.candidates[0].finish_reason
  assert chunks >= 1


def test_log_probs(client):
  client.models.generate_content(
      model=GEMINI_FLASH_2_0,
      contents='What is your name?',
      config={
          'logprobs': 2,
          'presence_penalty': 0.5,
          'frequency_penalty': 0.5,
          'response_logprobs': True,
      },
  )


def test_simple_config(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='What is your name?',
      config={
          'max_output_tokens': 300,
          'top_k': 2,
      },
  )
  assert response.text


def test_model_selection_config_dict(client):
  if not client.vertexai:
    return
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me a Taylor Swift lyric and explain its meaning.',
      config={
          'model_selection_config': {
              'feature_selection_preference': 'PRIORITIZE_COST'
          }
      },
  )
  assert response.text


def test_model_selection_config_pydantic(client):
  if not client.vertexai:
    return
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me a Taylor Swift lyric and explain its meaning.',
      config=types.GenerateContentConfig(
          model_selection_config=types.ModelSelectionConfig(
              feature_selection_preference=types.FeatureSelectionPreference.PRIORITIZE_QUALITY
          )
      ),
  )
  assert response.text


def test_sdk_logger_logs_warnings_once(client, caplog):
  from ... import types as types_module

  types_module._response_text_warning_logged = False

  caplog.set_level(logging.WARNING, logger='google_genai.types')

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Tell me a 50 word story about cheese.',
      config={
        'candidate_count': 2,
      }
  )
  assert response.text
  assert 'WARNING' in caplog.text
  assert 'there are 2 candidates' in caplog.text
  caplog_after_first_call = caplog.text
  assert len(caplog.records) == 1
  client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Tell me a 50 word story about cheese.',
      config={
        'candidate_count': 2,
      }
  )
  assert caplog.text == caplog_after_first_call
  assert len(caplog.records) == 1


def test_response_create_time_and_response_id(client):
  if client.vertexai:
    response = client.models.generate_content(
        model=GEMINI_FLASH_LATEST,
        contents='What is your name?',
        config={
            'max_output_tokens': 300,
            'top_k': 2,
        },
    )
    # create_time and response_id are not supported in mldev
    assert response.create_time
    assert response.response_id
    assert isinstance(response.create_time, datetime)


def test_safety_settings(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='What is your name?',
      config={
          'safety_settings': [{
              'category': 'HARM_CATEGORY_HATE_SPEECH',
              'threshold': 'BLOCK_ONLY_HIGH',
          }]
      },
  )
  assert response.text


def test_safety_settings_on_difference_stream(client):
  safety_settings = [
      {
          'category': 'HARM_CATEGORY_HATE_SPEECH',
          'threshold': 'BLOCK_ONLY_HIGH',
          'method': 'SEVERITY',
      },
      {
          'category': 'HARM_CATEGORY_DANGEROUS_CONTENT',
          'threshold': 'BLOCK_LOW_AND_ABOVE',
          'method': 'PROBABILITY',
      },
  ]
  if client._api_client.vertexai:
    for part in client.models.generate_content_stream(
        model=GEMINI_FLASH_LATEST,
        contents='What is your name?',
        config={
            'safety_settings': safety_settings,
        },
    ):
      pass
  else:
    with pytest.raises(ValueError) as e:
      for part in client.models.generate_content_stream(
          model=GEMINI_FLASH_LATEST,
          contents='What is your name?',
          config={
              'safety_settings': safety_settings,
          },
      ):
        pass
    assert 'method' in str(e)


def test_safety_settings_on_difference_stream_with_lower_enum(client):
  safety_settings = [
      {
          'category': 'harm_category_hate_speech',
          'threshold': 'block_only_high',
          'method': 'severity',
      },
      {
          'category': 'harm_category_dangerous_content',
          'threshold': 'block_low_and_above',
          'method': 'probability',
      },
  ]
  if client._api_client.vertexai:
    for part in client.models.generate_content_stream(
        model=GEMINI_FLASH_LATEST,
        contents='What is your name?',
        config={
            'safety_settings': safety_settings,
        },
    ):
      pass
  else:
    with pytest.raises(ValueError) as e:
      for part in client.models.generate_content_stream(
          model=GEMINI_FLASH_LATEST,
          contents='What is your name?',
          config={
              'safety_settings': safety_settings,
          },
      ):
        pass
    assert 'method' in str(e)


def test_pydantic_schema(client):
  class CountryInfo(BaseModel):
    # We need at least one test with `title` in properties in case the schema
    # edits go wrong.
    title: str
    population: int
    capital: str
    continent: str
    gdp: int
    official_language: str
    total_area_sq_mi: int

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information of the United States.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': CountryInfo,
      },
  )
  assert isinstance(response.parsed, CountryInfo)

def test_json_schema_fields(client):
  class UserRole(str, Enum):
    ADMIN = "admin"
    VIEWER = "viewer"
  class Address(BaseModel):
    street: str
    city: str
  class UserProfile(BaseModel):
    username: str = Field(description="User's unique name")
    age: Optional[int] = Field(ge=0, le=20)
    roles: Set[UserRole] = Field(min_items=1)
    contact: Union[Address, str]

    model_config = ConfigDict(
        title="User Schema", description="A user profile"
    )  # This is the title of the schema

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information of the United States.',
      config={
          'response_mime_type': 'application/json',
          'response_json_schema': UserProfile.model_json_schema(),
      },
  )
  print(response.parsed)
  assert response.parsed != None


def test_pydantic_schema_orders_properties(client):
  class Restaurant(BaseModel):
    name: str
    rating: int
    fun_fact: str

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information about a restaurant in Boston.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': Restaurant,
      },
  )
  response_text_json = json.loads(response.text)
  response_keys = list(response_text_json.keys())
  assert response_keys[0] == 'name'
  assert response_keys == list(Restaurant.model_fields.keys())


def test_pydantic_schema_with_default_value(client):
  class Restaurant(BaseModel):
    name: str
    rating: int = 0
    city: Optional[str] = 'New York'

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Can you recommend a restaurant for me?',
      config={
          'response_mime_type': 'application/json',
          'response_schema': Restaurant,
      },
  )
  assert isinstance(response.parsed, Restaurant)


def test_repeated_pydantic_schema(client):
  # This tests the defs handling on the pydantic side.
  class Person(BaseModel):
    name: str

  class Relationship(BaseModel):
    relationship: str
    person1: Person
    person2: Person

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Create a couple.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': Relationship,
      },
  )
  assert isinstance(response.parsed, Relationship)


def test_int_schema(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents="what's your favorite number?",
      config={
          'response_mime_type': 'application/json',
          'response_schema': int,
      },
  )
  assert isinstance(response.parsed, int)


def test_nested_list_of_int_schema(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents="Can you return two matrices, a 2x3 and a 3x4?",
      config={
          'response_mime_type': 'application/json',
          'response_schema': list[list[list[int]]],
      },
  )
  assert isinstance(response.parsed[0][0][0], int)


def test_literal_schema(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Which ice cream flavor should I order?',
      config={
          'response_mime_type': 'application/json',
          'response_schema': Literal['chocolate', 'vanilla', 'cookie dough'],
      },
  )

  allowed_values = ['chocolate', 'vanilla', 'cookie dough']
  assert isinstance(response.parsed, str)
  assert response.parsed in allowed_values


def test_literal_schema_with_non_string_types_raises(client):
  with pytest.raises(ValueError) as e:
    client.models.generate_content(
        model=GEMINI_FLASH_LATEST,
        contents='Which ice cream flavor should I order?',
        config={
            'response_mime_type': 'application/json',
            'response_schema': Literal['chocolate', 'vanilla', 1],
        },
    )
  assert 'validation error' in str(e)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason='| is not supported in Python 3.9',
)
def test_pydantic_schema_with_literal(client):
  class Movie(BaseModel):
    name: str
    genre: Literal['action', 'comedy', 'drama']

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information about the movie "Mean Girls"',
      config={
          'response_mime_type': 'application/json',
          'response_schema': Movie,
      },
  )
  assert isinstance(response.parsed, Movie)
  assert isinstance(response.parsed.genre, str)
  assert response.parsed.genre in ['action', 'comedy', 'drama']


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason='| is not supported in Python 3.9',
)
def test_pydantic_schema_with_single_value_literal(client):
  class Movie(BaseModel):
    name: str
    genre: Literal['action']

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information about the movie "The Matrix"',
      config={
          'response_mime_type': 'application/json',
          'response_schema': Movie,
      },
  )
  assert isinstance(response.parsed, Movie)
  assert response.parsed.genre == 'action'


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason='| is not supported in Python 3.9',
)
def test_pydantic_schema_with_none(client):
  class CountryInfo(BaseModel):
    name: str
    total_area_sq_mi: int | None = None

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information of the United States.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': CountryInfo,
      },
  )
  assert isinstance(response.parsed, CountryInfo)
  assert type(response.parsed.total_area_sq_mi) in [int, None]


def test_pydantic_schema_with_optional_none(client):
  class CountryInfo(BaseModel):
    name: str
    total_area_sq_mi: Optional[int] = None

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information of the United States but don\'t include the total area.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': CountryInfo,
      },
  )
  assert isinstance(response.parsed, CountryInfo)
  assert response.parsed.total_area_sq_mi is None


def test_pydantic_schema_from_json(client):
  class CountryInfo(BaseModel):
    name: str
    pupulation: int
    capital: str
    continent: str
    gdp: int
    official_language: str
    total_area_sq_mi: int

  schema = types.Schema.model_validate(CountryInfo.model_json_schema())

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information of the United States.',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=schema,
      ),
  )

  assert response.text


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason='| is not supported in Python 3.9',
)
def test_schema_with_union_type(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me a random number, either as an integers or written out as words.',
      config=types.GenerateContentConfig.model_validate(dict(
          response_mime_type='application/json',
          response_schema=int | str,
      ))
  )
  assert type(response.parsed) in (int, str)


def test_schema_with_union_type_all_py_versions(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents="Give me a random number, either an integer or a float.",
      config={
          'response_mime_type': 'application/json',
          'response_schema': Union[int, float],
      },
  )
  assert type(response.parsed) in (int, float)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason='| is not supported in Python 3.9',
)
def test_list_schema_with_union_type(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me a list of 5 random numbers, including some integers and some written out as words.',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=list[int | str],
      )
  )
  for item in response.parsed:
    assert isinstance(item, int) or isinstance(item, str)


def test_list_schema_with_union_type_all_py_versions(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me a list of 5 random numbers, including some integers and some written out as words.',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=list[Union[int, str]],
      )
  )
  for item in response.parsed:
    assert isinstance(item, int) or isinstance(item, str)


def test_pydantic_schema_with_optional_generic_alias(client):
  class CountryInfo(BaseModel):
    name: str
    population: int
    capital: str
    continent: str
    gdp: int
    official_languages: Optional[List[str]]
    total_area_sq_mi: int

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information of the United States.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': CountryInfo,
      },
  )
  assert isinstance(response.parsed, CountryInfo)
  assert isinstance(response.parsed.official_languages, list) or response.parsed.official_languages is None


def test_pydantic_schema_with_optional_pydantic(client):
  class TestPerson(BaseModel):
    first_name: Optional[str] = Field(
        description='First name of the person', default=None
    )
    last_name: Optional[str] = Field(
        description='Last name of the person', default=None
    )

  class TestDocument(BaseModel):
    case_number: Optional[str] = Field(
        description='Case number assigned to the claim', default=None
    )
    filed_by: Optional[TestPerson] = Field(
        description='Name of the party that filed or submitted the statement',
        default=None,
    )

  test_prompt = """
  Carefully examine the following document and extract the metadata.
  Be sure to include the party that filed the document.

  Document Text:
  --------------
  Case Number: 20-12345
  File by: John Doe
  """

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents=test_prompt,
      config={
          'response_mime_type': 'application/json',
          'response_schema': TestDocument,
      },
  )
  assert isinstance(response.parsed, TestDocument)
  assert isinstance(response.parsed.filed_by, TestPerson)


def test_list_of_pydantic_schema(client):
  class CountryInfo(BaseModel):
    name: str
    population: int
    capital: str
    continent: str
    gdp: int
    official_language: str
    total_area_sq_mi: int

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information for the United States, Canada, and Mexico.',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=list[CountryInfo],
      )
  )
  assert isinstance(response.parsed, list)
  assert len(response.parsed) == 3
  assert isinstance(response.parsed[0], CountryInfo)


def test_nested_list_of_pydantic_schema(client):
  class Recipe(BaseModel):
    name: str
    cook_time: str

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents="I\'m writing three recipe books, one each for United States, Canada, and Mexico. "
               "Can you give some recipe ideas, at least 2 per book?",
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=list[list[Recipe]],
      )
  )
  assert isinstance(response.parsed, list)
  assert len(response.parsed) == 3
  assert isinstance(response.parsed[0][0], Recipe)


def test_list_of_pydantic_schema_with_dict_config(client):
  class CountryInfo(BaseModel):
    name: str
    population: int
    capital: str
    continent: str
    gdp: int
    official_language: str
    total_area_sq_mi: int

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information for the United States, Canada, and Mexico.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': list[CountryInfo],
      }
  )
  assert isinstance(response.parsed, list)
  assert len(response.parsed) == 3
  assert isinstance(response.parsed[0], CountryInfo)


def test_pydantic_schema_with_nested_class(client):
  class CurrencyInfo(BaseModel):
    name: str

  class CountryInfo(BaseModel):
    name: str
    currency: CurrencyInfo

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information for the United States',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=CountryInfo,
      )
  )
  assert isinstance(response.parsed, CountryInfo)
  assert isinstance(response.parsed.currency, CurrencyInfo)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason='| is not supported in Python 3.9',
)
def test_pydantic_schema_with_union_type(client):

  class CountryInfo(BaseModel):
    name: str
    restaurants_per_capita: int | float

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information for the United States',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=CountryInfo,
      )
  )
  assert isinstance(response.parsed, CountryInfo)
  assert type(response.parsed.restaurants_per_capita) in (int, float)


def test_pydantic_schema_with_union_type_all_py_versions(client):

  class CountryInfo(BaseModel):
    name: str
    restaurants_per_capita: Union[int, float]

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information for the United States',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=CountryInfo,
      )
  )
  assert isinstance(response.parsed, CountryInfo)
  assert type(response.parsed.restaurants_per_capita) in (int, float)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason='| is not supported in Python 3.9',
)
def test_union_of_pydantic_schema(client):

  class SongLyric(BaseModel):
    song_name: str
    lyric: str
    artist: str

  class FunFact(BaseModel):
    fun_fact: str

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Can you give me a Taylor Swift song lyric or a fun fact?',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=SongLyric | FunFact,
      )
  )
  assert type(response.parsed) in (SongLyric, FunFact)


def test_union_of_pydantic_schema_all_py_versions(client):

  class SongLyric(BaseModel):
    song_name: str
    lyric: str
    artist: str

  class FunFact(BaseModel):
    fun_fact: str

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Can you give me a Taylor Swift song lyric or a fun fact?',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=Union[SongLyric, FunFact],
      )
  )
  assert type(response.parsed) in (SongLyric, FunFact)


def test_pydantic_schema_with_nested_enum(client):
  class Continent(Enum):
    ASIA = 'Asia'
    AFRICA = 'Africa'
    ANTARCTICA = 'Antarctica'
    EUROPE = 'Europe'
    NORTH_AMERICA = 'North America'
    SOUTH_AMERICA = 'South America'
    AUSTRALIA = 'Australia'

  class CountryInfo(BaseModel):
    name: str
    continent: Continent

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information for the United States',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=CountryInfo,
      )
  )
  assert isinstance(response.parsed, CountryInfo)
  assert isinstance(response.parsed.continent, Continent)


def test_pydantic_schema_with_nested_list_class(client):
  class CurrencyInfo(BaseModel):
    name: str

  class CountryInfo(BaseModel):
    name: str
    currency: list[CurrencyInfo]

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information for the United States.',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=CountryInfo,
      )
  )
  assert isinstance(response.parsed, CountryInfo)
  assert isinstance(response.parsed.currency[0], CurrencyInfo)


def test_list_of_pydantic_schema_with_nested_class(client):
  class CurrencyInfo(BaseModel):
    name: str
    code: str
    symbol: str

  class CountryInfo(BaseModel):
    name: str
    population: int
    capital: str
    continent: str
    gdp: int
    official_language: str
    total_area_sq_mi: int
    currency: CurrencyInfo

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information for the United States, Canada, and Mexico.',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=list[CountryInfo],
      )
  )
  assert isinstance(response.parsed, list)
  assert isinstance(response.parsed[0], CountryInfo)
  assert isinstance(response.parsed[0].currency, CurrencyInfo)


def test_list_of_pydantic_schema_with_nested_list_class(client):
  class CurrencyInfo(BaseModel):
    name: str
    code: str
    symbol: str

  class CountryInfo(BaseModel):
    name: str
    population: int
    capital: str
    continent: str
    gdp: int
    official_language: str
    total_area_sq_mi: int
    currency: list[CurrencyInfo]

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information for the United States, Canada, and Mexico.',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=list[CountryInfo],
      )
  )
  assert isinstance(response.parsed, list)
  assert isinstance(response.parsed[0], CountryInfo)
  assert isinstance(response.parsed[0].currency, list)
  assert isinstance(response.parsed[0].currency[0], CurrencyInfo)


def test_response_schema_with_dict_of_pydantic_schema(client):
  class CountryInfo(BaseModel):
    population: int
    capital: str
    continent: str
    gdp: int
    official_language: str
    total_area_sq_mi: int

  if not client.vertexai:
    with pytest.raises(ValueError) as e:
      client.models.generate_content(
          model=GEMINI_FLASH_LATEST,
          contents='Give me information for the United States, Canada, and Mexico.',
          config=types.GenerateContentConfig(
              response_mime_type='application/json',
              response_schema=dict[str, CountryInfo],
          )
      )
  else:
    response = client.models.generate_content(
        model=GEMINI_FLASH_LATEST,
        contents='Give me information for the United States, Canada, and Mexico.',
        config=types.GenerateContentConfig(
            response_mime_type='application/json',
            response_schema=dict[str, CountryInfo],
        )
    )
    assert response.text


def test_schema_with_unsupported_type_raises(client):
  with pytest.raises(ValueError) as e:
    client.models.generate_content(
        model=GEMINI_FLASH_LATEST,
        contents='Give me information for the United States, Canada, and Mexico.',
        config=types.GenerateContentConfig(
            response_mime_type='application/json',
            response_schema=types.Schema(),
        )
    )
  assert 'Unsupported schema type' in str(e)


def test_enum_schema_with_enum_mime_type(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_2_0,
      contents='What instrument plays multiple notes at once?',
      config={
          'response_mime_type': 'text/x.enum',
          'response_schema': InstrumentEnum,
      },
  )

  instrument_values = {member.value for member in InstrumentEnum}

  assert response.text in instrument_values
  assert isinstance(response.parsed, InstrumentEnum)


def test_list_of_enum_schema_with_enum_mime_type(client):
  with pytest.raises(errors.ClientError) as e:
    client.models.generate_content(
        model=GEMINI_FLASH_2_0,
        contents='What instrument plays single note at once?',
        config={
            'response_mime_type': 'text/x.enum',
            'response_schema': list[InstrumentEnum],
        },
    )
  assert '400' in str(e)


def test_list_of_enum_schema_with_json_mime_type(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='What instrument plays single note at once?',
      config={
          'response_mime_type': 'application/json',
          'response_schema': list[InstrumentEnum],
      },
  )

  assert isinstance(response.parsed, list)
  assert response.parsed
  for item in response.parsed:
    assert isinstance(item, InstrumentEnum)


def test_optional_enum_in_pydantic_schema_with_json_mime_type(client):
  class InstrumentInfo(BaseModel):
    instrument: Optional[InstrumentEnum]
    fun_fact: str

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='What instrument plays single note at once? Include the name of the instrument in your response.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': InstrumentInfo,
      },
  )

  assert isinstance(response.parsed, InstrumentInfo)
  assert isinstance(response.parsed.instrument, InstrumentEnum)


def test_enum_schema_with_json_mime_type(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='What instrument plays multiple notes at once?',
      config={
          'response_mime_type': 'application/json',
          'response_schema': InstrumentEnum,
      },
  )
  # "application/json" returns response in double quotes.
  removed_quotes = response.text.replace('"', '')
  instrument_values = {member.value for member in InstrumentEnum}

  assert removed_quotes in instrument_values
  assert isinstance(response.parsed, InstrumentEnum)


def test_non_string_enum_schema_with_enum_mime_type(client):
  class IntegerEnum(Enum):
    PERCUSSION = 1
    STRING = 2
    WOODWIND = 3
    BRASS = 4
    KEYBOARD = 5

  response =client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='What instrument plays multiple notes at once?',
      config={
          'response_mime_type': 'text/x.enum',
          'response_schema': IntegerEnum,
      },
  )

  instrument_values = {str(member.value) for member in IntegerEnum}

  assert response.text in instrument_values


def test_json_schema(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information of the United States.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': {
              'required': [
                  'name',
                  'population',
                  'capital',
                  'continent',
                  'gdp',
                  'official_language',
                  'total_area_sq_mi',
              ],
              'properties': {
                  'name': {'type': 'STRING'},
                  'population': {'type': 'INTEGER'},
                  'capital': {'type': 'STRING'},
                  'continent': {'type': 'STRING'},
                  'gdp': {'type': 'INTEGER'},
                  'official_language': {'type': 'STRING'},
                  'total_area_sq_mi': {'type': 'INTEGER'},
              },
              'type': 'OBJECT',
          },
      },
  )
  assert isinstance(response.parsed, dict)


def test_json_schema_with_lower_enum(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information of the United States.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': {
              'required': [
                  'name',
                  'pupulation',
                  'capital',
                  'continent',
                  'gdp',
                  'official_language',
                  'total_area_sq_mi',
              ],
              'properties': {
                  'name': {'type': 'string'},
                  'pupulation': {'type': 'integer'},
                  'capital': {'type': 'string'},
                  'continent': {'type': 'string'},
                  'gdp': {'type': 'integer'},
                  'official_language': {'type': 'string'},
                  'total_area_sq_mi': {'type': 'integer'},
              },
              'type': 'OBJECT',
          },
      },
  )
  assert isinstance(response.parsed, dict)


def test_json_schema_with_any_of(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me a fruit basket.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': {
              'type': 'OBJECT',
              'title': 'Fruit Basket',
              'description': 'A structured representation of a fruit basket',
              'required': ['fruit'],
              'properties': {
                  'fruit': {
                      'type': 'ARRAY',
                      'description': (
                          'An ordered list of the fruit in the basket'
                      ),
                      'items': {
                          'description': 'A piece of fruit',
                          'any_of': [
                              {
                                  'title': 'Apple',
                                  'description': 'Describes an apple',
                                  'type': 'OBJECT',
                                  'properties': {
                                      'type': {
                                          'type': 'STRING',
                                          'description': "Always 'apple'",
                                      },
                                      'color': {
                                          'type': 'STRING',
                                          'description': (
                                              'The color of the apple (e.g.,'
                                              " 'red')"
                                          ),
                                      },
                                  },
                                  'property_ordering': ['type', 'color'],
                                  'required': ['type', 'color'],
                              },
                              {
                                  'title': 'Orange',
                                  'description': 'Describes an orange',
                                  'type': 'OBJECT',
                                  'properties': {
                                      'type': {
                                          'type': 'STRING',
                                          'description': "Always 'orange'",
                                      },
                                      'size': {
                                          'type': 'STRING',
                                          'description': (
                                              'The size of the orange (e.g.,'
                                              " 'medium')"
                                          ),
                                      },
                                  },
                                  'property_ordering': ['type', 'size'],
                                  'required': ['type', 'size'],
                              },
                          ],
                      },
                  }
              },
          },
      },
  )
  assert isinstance(response.parsed, dict)
  assert 'fruit' in response.parsed
  assert isinstance(response.parsed['fruit'], list)
  assert 'type' in response.parsed['fruit'][0]


def test_schema_with_any_of(client):
  response_schema=types.Schema(
      type=types.Type.OBJECT,
      title='Fruit Basket',
      description='A structured representation of a fruit basket',
      properties={
          'fruit': types.Schema(
              type=types.Type.ARRAY,
              description='An ordered list of the fruit in the basket',
              items=types.Schema(
                  any_of=[
                      types.Schema(
                          title='Apple',
                          description='Describes an apple',
                          type=types.Type.OBJECT,
                          properties={
                              'type': types.Schema(type=types.Type.STRING, description='Always "apple"'),
                              'variety': types.Schema(
                                  type=types.Type.STRING,
                                  description='The variety of apple (e.g., "Granny Smith")',
                              ),
                          },
                          property_ordering=['type', 'variety'],
                          required=['type', 'variety'],
                      ),
                      types.Schema(
                          title='Orange',
                          description='Describes an orange',
                          type=types.Type.OBJECT,
                          properties={
                              'type': types.Schema(type=types.Type.STRING, description='Always "orange"'),
                              'variety': types.Schema(
                                  type=types.Type.STRING,
                                  description='The variety of orange (e.g.,"Navel orange")',
                              ),
                          },
                          property_ordering=['type', 'variety'],
                          required=['type', 'variety'],
                      ),
                  ],
              ),
          ),
      },
      required=['fruit'],
  )
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Give me a fruit basket.',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=response_schema,
      ),
  )
  assert isinstance(response.parsed, dict)
  assert 'fruit' in response.parsed
  assert isinstance(response.parsed['fruit'], list)
  assert 'type' in response.parsed['fruit'][0]


def test_replicated_voice_config(client):
  with pytest_helper.exception_if_vertex(client, errors.ClientError):
      client.models.generate_content(
          model='gemini-2.5-flash-preview-tts-voice-replication-rev22-2025-10-28',
          contents=t.t_contents(
              'Produce a speech response saying "Cheese"'
          ),
          config=types.GenerateContentConfig(
              response_modalities=['audio'],
              speech_config=types.SpeechConfig(
                  voice_config=types.VoiceConfig(
                      replicated_voice_config=types.ReplicatedVoiceConfig(
                          voice_sample_audio=audio_bytes,
                          mime_type='audio/wav',
                      )
                  )
              ),
          ),
      )


def test_json_schema_with_streaming(client):

  response = client.models.generate_content_stream(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information of the United States.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': {
              'properties': {
                  'name': {'type': 'STRING'},
                  'population': {'type': 'INTEGER'},
                  'capital': {'type': 'STRING'},
                  'continent': {'type': 'STRING'},
                  'gdp': {'type': 'INTEGER'},
                  'official_language': {'type': 'STRING'},
                  'total_area_sq_mi': {'type': 'INTEGER'},
              },
              'type': 'OBJECT',
          },
      },
  )

  for r in response:
    parts = r.parts
    for p in parts:
      assert p.text


def test_pydantic_schema_with_streaming(client):

  class CountryInfo(BaseModel):
    name: str
    population: int
    capital: str
    continent: str
    gdp: int
    official_language: str
    total_area_sq_mi: int

  response = client.models.generate_content_stream(
      model=GEMINI_FLASH_LATEST,
      contents='Give me information of the United States.',
      config={
          'response_mime_type': 'application/json',
          'response_schema': CountryInfo
      },
  )

  for r in response:
    parts = r.parts
    for p in parts:
      assert p.text


def test_schema_from_json(client):

  class Foo(BaseModel):
    bar: str
    baz: int
    qux: list[str]

  schema = types.Schema.model_validate(Foo.model_json_schema())

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Fill in the Foo.',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=schema
      ),
  )

  assert response.text


def test_schema_from_model_schema(client):

  class Foo(BaseModel):
    bar: str
    baz: int
    qux: list[str]

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='Fill in the Foo.',
      config=types.GenerateContentConfig(
          response_mime_type='application/json',
          response_schema=Foo.model_json_schema(),
      ),
  )

  response.text


def test_schema_with_additional_properties(client):

  class Foo(BaseModel):
    bar: str
    baz: int
    qux: dict[str, str]

  if client.vertexai:
    response = client.models.generate_content(
        model=GEMINI_FLASH_LATEST,
        contents='What is your name?',
        config=types.GenerateContentConfig(
            response_mime_type='application/json',
            response_schema=Foo,
        ),
    )
    assert response.text
  else:
    with pytest.raises(ValueError) as e:
      client.models.generate_content(
          model=GEMINI_FLASH_LATEST,
          contents='What is your name?',
          config=types.GenerateContentConfig(
              response_mime_type='application/json',
              response_schema=Foo,
          ),
      )
    assert 'additionalProperties is not supported in the Gemini API.' in str(e)


def test_function(client):
  def get_weather(city: str) -> str:
    """Returns the weather in a city."""
    return f'The weather in {city} is sunny and 100 degrees.'

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents=(
          'What is the weather like in Sunnyvale? Answer in very short'
          ' sentence.'
      ),
      config={
          'tools': [get_weather],
      },
  )
  assert '100' in response.text


def test_invalid_input_without_transformer(client):
  with pytest.raises(ValidationError) as e:
    client.models.generate_content(
        model=GEMINI_FLASH_LATEST,
        contents='What is your name',
        config={
            'input_that_does_not_exist': 'what_ever_value',
        },
    )
  assert 'input_that_does_not_exist' in str(e)
  assert 'Extra inputs are not permitted' in str(e)


def test_invalid_input_with_transformer_dict(client):
  with pytest.raises(ValidationError) as e:
    client.models.generate_content(
        model=GEMINI_FLASH_LATEST,
        contents={'invalid_key': 'invalid_value'},
    )
  assert 'invalid_key' in str(e.value)


def test_invalid_input_with_transformer_list(client):
  with pytest.raises(ValidationError) as e:
    client.models.generate_content(
        model=GEMINI_FLASH_LATEST,
        contents=[{'invalid_key': 'invalid_value'}],
    )
  assert 'invalid_key' in str(e.value)


def test_invalid_input_for_simple_parameter(client):
  with pytest.raises(ValidationError) as e:
    client.models.generate_content(
        model=5,
        contents='What is your name?',
    )
  assert 'model' in str(e)


def test_catch_stack_trace_in_error_handling(client):
  try:
    client.models.generate_content(
        model=GEMINI_FLASH_LATEST,
        contents='What is your name?',
        config={'response_modalities': ['AUDIO']},
    )
  except errors.ClientError as e:
    # Note that the stack trace is truncated in replay file, therefore this is
    # the best we can do in testing error handling. In api mode, the stack trace
    # is:
    # {
    #     'error': {
    #         'code': 400,
    #         'message': 'Multi-modal output is not supported.',
    #         'status': 'INVALID_ARGUMENT',
    #         'details': [{
    #             '@type': 'type.googleapis.com/google.rpc.DebugInfo',
    #             'detail': '[ORIGINAL ERROR] generic::invalid_argument: '
    #                      'Multi-modal output is not supported. '
    #                      '[google.rpc.error_details_ext] '
    #                      '{ message: "Multi-modal output is not supported." }'
    #         }]
    #     }
    # }
    if 'error' in e.details:
      details = e.details['error']
    else:
      details = e.details
    assert details['code'] == 400
    assert details['status'] == 'INVALID_ARGUMENT'


def test_multiple_strings(client):
  class SummaryResponses(BaseModel):
    summary: str
    person: str

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents=[
          "Summarize Shakespeare's life work in a few sentences",
          "Summarize Hemingway's life work",
      ],
      config={
          'response_mime_type': 'application/json',
          'response_schema': list[SummaryResponses],
      },
  )

  assert 'Shakespeare' in response.text
  assert 'Hemingway' in response.text
  assert 'Shakespeare' in response.parsed[0].person
  assert 'Hemingway' in response.parsed[1].person


def test_multiple_parts(client):
  class SummaryResponses(BaseModel):
    summary: str
    person: str

  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents=[
          types.Part(
              text="Summarize Shakespeare's life work in a few sentences"
          ),
          types.Part(text="Summarize Hemingway's life work"),
      ],
      config={
          'response_mime_type': 'application/json',
          'response_schema': list[SummaryResponses],
      },
  )

  assert 'Shakespeare' in response.text
  assert 'Hemingway' in response.text
  assert 'Shakespeare' in  response.parsed[0].person
  assert 'Hemingway' in response.parsed[1].person


def test_multiple_function_calls(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents=[
          'What is the weather in Boston?',
          'What is the stock price of GOOG?',
          types.Part.from_function_call(
              name='get_weather',
              args={'location': 'Boston'},
          ),
          types.Part.from_function_call(
              name='get_stock_price',
              args={'symbol': 'GOOG'},
          ),
          types.Part.from_function_response(
              name='get_weather',
              response={'response': 'It is sunny and 100 degrees.'},
          ),
          types.Part.from_function_response(
              name='get_stock_price',
              response={'response': 'The stock price is $100.'},
          ),
      ],
      config=types.GenerateContentConfig(
          tools=[
              types.Tool(
                  function_declarations=[
                      types.FunctionDeclaration(
                          name='get_weather',
                          description='Get the weather in a city.',
                          parameters=types.Schema(
                              type=types.Type.OBJECT,
                              properties={
                                  'location': types.Schema(
                                      type=types.Type.STRING
                                  )
                              },
                          ),
                      ),
                      types.FunctionDeclaration(
                          name='get_stock_price',
                          description='Get the stock price of a symbol.',
                          parameters=types.Schema(
                              type=types.Type.OBJECT,
                              properties={
                                  'symbol': types.Schema(
                                      type=types.Type.STRING
                                  )
                              },
                          ),
                      ),
                  ]
              ),
          ]
      ),
  )

  assert 'Boston' in response.text
  assert 'sunny' in response.text
  assert '100 degrees' in response.text
  assert '$100' in response.text


def test_usage_metadata_part_types(client):
  contents = [
      'Hello world.',
      types.Part.from_bytes(
          data=image_bytes,
          mime_type='image/png',
      ),
  ]

  response = client.models.generate_content(
      model=GEMINI_FLASH_2_0, contents=contents
  )
  usage_metadata = response.usage_metadata

  assert usage_metadata.candidates_token_count
  assert usage_metadata.candidates_tokens_details
  modalities = sorted(
      [d.modality.name for d in usage_metadata.candidates_tokens_details]
  )
  assert modalities == ['TEXT']
  assert isinstance(
      usage_metadata.candidates_tokens_details[0].modality, types.MediaModality)

  assert usage_metadata.prompt_token_count
  assert usage_metadata.prompt_tokens_details
  modalities = sorted(
      [d.modality.name for d in usage_metadata.prompt_tokens_details]
  )
  assert modalities == ['IMAGE', 'TEXT']


def test_error_handling_stream(client):
  if client.vertexai:
    return

  try:
    for chunk in client.models.generate_content_stream(
        model=GEMINI_FLASH_IMAGE_LATEST,
        contents=[
            types.Content(
                role='user',
                parts=[
                    types.Part.from_bytes(
                        data=image_bytes, mime_type='image/png'
                    ),
                    types.Part.from_text(text='Make sky more beautiful.'),
                ],
            ),
        ],
        config=types.GenerateContentConfig(
            response_mime_type='text/plain',
            response_modalities=['IMAGE', 'TEXT'],
            system_instruction='make the sky more beautiful.',
        ),
    ):
      continue

  except errors.ClientError as e:
    assert (
        e.message
        == 'Developer instruction is not enabled for'
        ' models/gemini-2.5-flash-image'
    )


def test_error_handling_unary(client):
  if client.vertexai:
    return

  try:
    client.models.generate_content(
        model=GEMINI_FLASH_IMAGE_LATEST,
        contents=[
            types.Content(
                role='user',
                parts=[
                    types.Part.from_bytes(
                        data=image_bytes, mime_type='image/png'
                    ),
                    types.Part.from_text(text='Make sky more beautiful.'),
                ],
            ),
        ],
        config=types.GenerateContentConfig(
            response_mime_type='text/plain',
            response_modalities=['IMAGE', 'TEXT'],
            system_instruction='make the sky more beautiful.',
        ),
    )

  except errors.ClientError as e:
    assert (
        e.message
        == 'Developer instruction is not enabled for'
        ' models/gemini-2.5-flash-image'
    )


def test_provisioned_output_dedicated(client):
  response = client.models.generate_content(
      model=GEMINI_FLASH_LATEST,
      contents='What is 1 + 1?',
      config=types.GenerateContentConfig(
          http_options={'headers': {'X-Vertex-AI-LLM-Request-Type': 'dedicated'}}
      ),
  )
  if client.vertexai:
    assert response.usage_metadata.traffic_type == types.TrafficType.PROVISIONED_THROUGHPUT
  else:
    assert not response.usage_metadata.traffic_type


@pytest.mark.asyncio
async def test_error_handling_unary_async(client):
  if client.vertexai:
    return

  try:
    await client.aio.models.generate_content(
        model=GEMINI_FLASH_IMAGE_LATEST,
        contents=[
            types.Content(
                role='user',
                parts=[
                    types.Part.from_bytes(
                        data=image_bytes, mime_type='image/png'
                    ),
                    types.Part.from_text(text='Make sky more beautiful.'),
                ],
            ),
        ],
        config=types.GenerateContentConfig(
            response_mime_type='text/plain',
            response_modalities=['IMAGE', 'TEXT'],
            system_instruction='make the sky more beautiful.',
        ),
    )

  except errors.ClientError as e:
    assert (
        e.message
        == 'Developer instruction is not enabled for'
        ' models/gemini-2.5-flash-image'
    )


@pytest.mark.asyncio
async def test_error_handling_stream_async(client):
  if client.vertexai:
    return

  try:
    async for part in await client.aio.models.generate_content_stream(
        model=GEMINI_FLASH_IMAGE_LATEST,
        contents=[
            types.Content(
                role='user',
                parts=[
                    types.Part.from_bytes(
                        data=image_bytes, mime_type='image/png'
                    ),
                    types.Part.from_text(text='Make sky more beautiful.'),
                ],
            ),
        ],
        config=types.GenerateContentConfig(
            response_mime_type='text/plain',
            response_modalities=['IMAGE', 'TEXT'],
            system_instruction='make the sky more beautiful.',
        ),
    ):
      continue

  except errors.ClientError as e:
    assert ('Developer instruction is not enabled' in e.message)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests for batches.create() with file source."""

import pytest

from ... import types
from .. import pytest_helper


_GEMINI_MODEL = 'gemini-2.5-flash'
_DISPLAY_NAME = 'test_batch'

_MLDEV_GEMINI_MODEL = 'gemini-2.5-flash'
_FILE_NAME = 'files/s0pa54alni6w'

# All tests will be run for both Vertex and MLDev.
test_table: list[pytest_helper.TestTableItem] = [
    pytest_helper.TestTableItem(
        name='test_union_generate_content_with_file',
        parameters=types._CreateBatchJobParameters(
            model=_MLDEV_GEMINI_MODEL,
            src={'file_name': _FILE_NAME},
            config={
                'display_name': _DISPLAY_NAME,
            },
        ),
        exception_if_vertex='not supported',
        has_union=True,
    ),
    pytest_helper.TestTableItem(
        name='test_generate_content_with_file',
        parameters=types._CreateBatchJobParameters(
            model=_MLDEV_GEMINI_MODEL,
            src={'file_name': _FILE_NAME},
            config={
                'display_name': _DISPLAY_NAME,
            },
        ),
        exception_if_vertex='not supported',
    ),
]

pytestmark = [
    pytest.mark.usefixtures('mock_timestamped_unique_name'),
    pytest_helper.setup(
        file=__file__,
        globals_for_file=globals(),
        test_method='batches.create',
        test_table=test_table,
    ),
]


@pytest.mark.asyncio
async def test_async_create(client):
  with pytest_helper.exception_if_vertex(client, ValueError):
    batch_job = await client.aio.batches.create(
        model=_MLDEV_GEMINI_MODEL,
        src={'file_name': _FILE_NAME},
        config={
            'display_name': _DISPLAY_NAME,
        },
    )
    assert batch_job.name.startswith('batches/')
    assert (
        batch_job.model == 'models/' + _MLDEV_GEMINI_MODEL
    )  # Converted to Gemini full name.



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for generate_images."""

from .... import types
from ... import pytest_helper

IMAGEN_MODEL_LATEST = 'imagen-4.0-generate-001'

test_table: list[pytest_helper.TestTableItem] = [
    pytest_helper.TestTableItem(
        name='test_simple_prompt',
        parameters=types._GenerateImagesParameters(
            model=IMAGEN_MODEL_LATEST,
            prompt='Red skateboard',
            config=types.GenerateImagesConfig(
                number_of_images=1,
                output_mime_type='image/jpeg',
            ),
        ),
    ),
]
pytestmark = pytest_helper.setup(
    file=__file__,
    globals_for_file=globals(),
    test_method='models.generate_images',
    test_table=test_table,
)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests for live.py."""
import json
import os
from unittest import mock

import pytest
from websockets import client

from ... import client as gl_client
from ... import live
from ... import types
from .. import pytest_helper


IMAGE_FILE_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '../data/google.jpg')
)


def mock_api_client(vertexai=False):
  api_client = mock.MagicMock(spec=gl_client.BaseApiClient)
  api_client.api_key = 'TEST_API_KEY'
  api_client._host = lambda: 'test_host'
  api_client._http_options = {'headers': {}}  # Ensure headers exist
  api_client.vertexai = vertexai
  api_client._api_client = api_client
  return api_client


@pytest.fixture
def mock_websocket():
  websocket = mock.AsyncMock(spec=client.ClientConnection)
  websocket.send = mock.AsyncMock()
  websocket.recv = mock.AsyncMock(
      return_value='{"serverContent": {"turnComplete": true}}'
  )  # Default response
  websocket.close = mock.AsyncMock()
  return websocket


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_send_media_blob_dict(mock_websocket, vertexai):
  session = live.AsyncSession(
      api_client=mock_api_client(vertexai=vertexai), websocket=mock_websocket
  )
  content = {'data': bytes([0, 0, 0, 0, 0, 0]), 'mime_type': 'audio/pcm'}

  await session.send_realtime_input(media=content)
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'realtime_input' in sent_data

  assert sent_data['realtime_input']['mediaChunks'][0]['data'] == 'AAAAAAAA'
  assert pytest_helper.get_value_ignore_key_case(
      sent_data['realtime_input']['mediaChunks'][0], 'mime_type'
  ) == 'audio/pcm'


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_send_media_blob(mock_websocket, vertexai):
  session = live.AsyncSession(
      api_client=mock_api_client(vertexai=vertexai), websocket=mock_websocket
  )
  content = types.Blob(data=bytes([0, 0, 0, 0, 0, 0]), mime_type='audio/pcm')

  await session.send_realtime_input(media=content)
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'realtime_input' in sent_data

  assert sent_data['realtime_input']['mediaChunks'][0]['data'] == 'AAAAAAAA'
  assert pytest_helper.get_value_ignore_key_case(
      sent_data['realtime_input']['mediaChunks'][0], 'mime_type'
  ) == 'audio/pcm'


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_send_media_image(mock_websocket, vertexai, image_jpeg):
  session = live.AsyncSession(
      api_client=mock_api_client(vertexai=vertexai), websocket=mock_websocket
  )

  await session.send_realtime_input(media=image_jpeg)
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'realtime_input' in sent_data

  assert pytest_helper.get_value_ignore_key_case(
      sent_data['realtime_input']['mediaChunks'][0], 'mime_type'
  ) == 'image/jpeg'


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.parametrize(
    'content',
    [
        {'data': bytes([0, 0, 0, 0, 0, 0]), 'mime_type': 'audio/pcm'},
        types.Blob(data=bytes([0, 0, 0, 0, 0, 0]), mime_type='audio/pcm'),
    ],
)
@pytest.mark.asyncio
async def test_send_audio(mock_websocket, vertexai, content):
  api_client = mock_api_client(vertexai=vertexai)
  session = live.AsyncSession(api_client=api_client, websocket=mock_websocket)

  await session.send_realtime_input(audio=content)
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'realtime_input' in sent_data

  assert sent_data['realtime_input']['audio']['data'] == 'AAAAAAAA'
  assert pytest_helper.get_value_ignore_key_case(
      sent_data['realtime_input']['audio'], 'mime_type'
  ) == 'audio/pcm'


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_send_bad_audio_blob(mock_websocket, vertexai):
  api_client = mock_api_client(vertexai=vertexai)
  session = live.AsyncSession(api_client=api_client, websocket=mock_websocket)
  content = types.Blob(data=bytes([0, 0, 0, 0, 0, 0]), mime_type='image/png')

  with pytest.raises(ValueError, match='.*Unsupported mime type.*'):
    await session.send_realtime_input(audio=content)


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_send_bad_video_blob(mock_websocket, vertexai):
  api_client = mock_api_client(vertexai=vertexai)
  session = live.AsyncSession(api_client=api_client, websocket=mock_websocket)
  content = types.Blob(data=bytes([0, 0, 0, 0, 0, 0]), mime_type='audio/pcm')

  with pytest.raises(ValueError, match='.*Unsupported mime type.*'):
    await session.send_realtime_input(video=content)


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_send_audio_stream_end(mock_websocket, vertexai):
  session = live.AsyncSession(
      api_client=mock_api_client(vertexai=vertexai), websocket=mock_websocket
  )

  await session.send_realtime_input(audio_stream_end=True)
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'realtime_input' in sent_data

  assert sent_data['realtime_input']['audioStreamEnd'] == True


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.parametrize(
    'content',
    [
        {'data': bytes([0, 0, 0, 0, 0, 0]), 'mime_type': 'image/png'},
        types.Blob(data=bytes([0, 0, 0, 0, 0, 0]), mime_type='image/png'),
    ],
)
@pytest.mark.asyncio
async def test_send_video(mock_websocket, vertexai, content):
  api_client = mock_api_client(vertexai=vertexai)
  session = live.AsyncSession(api_client=api_client, websocket=mock_websocket)

  await session.send_realtime_input(video=content)
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'realtime_input' in sent_data

  assert sent_data['realtime_input']['video']['data'] == 'AAAAAAAA'
  assert pytest_helper.get_value_ignore_key_case(
      sent_data['realtime_input']['video'], 'mime_type'
  ) == 'image/png'


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_send_video_image(mock_websocket, vertexai, image_jpeg):
  api_client = mock_api_client(vertexai=vertexai)
  session = live.AsyncSession(api_client=api_client, websocket=mock_websocket)

  await session.send_realtime_input(video=image_jpeg)
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'realtime_input' in sent_data

  assert pytest_helper.get_value_ignore_key_case(
      sent_data['realtime_input']['video'], 'mime_type'
  ) == 'image/jpeg'


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_send_text(mock_websocket, vertexai):
  api_client = mock_api_client(vertexai=vertexai)
  session = live.AsyncSession(api_client=api_client, websocket=mock_websocket)

  await session.send_realtime_input(text='Hello?')
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'realtime_input' in sent_data

  assert sent_data['realtime_input']['text'] == 'Hello?'


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.parametrize('activity', [{}, types.ActivityStart()])
@pytest.mark.asyncio
async def test_send_activity_start(mock_websocket, vertexai, activity):
  session = live.AsyncSession(
      api_client=mock_api_client(vertexai=vertexai), websocket=mock_websocket
  )

  await session.send_realtime_input(activity_start=activity)
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'realtime_input' in sent_data

  assert sent_data['realtime_input']['activityStart'] == {}


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.parametrize('activity', [{}, types.ActivityEnd()])
@pytest.mark.asyncio
async def test_send_activity_end(mock_websocket, vertexai, activity):
  session = live.AsyncSession(
      api_client=mock_api_client(vertexai=vertexai), websocket=mock_websocket
  )

  await session.send_realtime_input(activity_end=activity)
  mock_websocket.send.assert_called_once()
  sent_data = json.loads(mock_websocket.send.call_args[0][0])
  assert 'realtime_input' in sent_data

  assert sent_data['realtime_input']['activityEnd'] == {}


@pytest.mark.parametrize('vertexai', [True, False])
@pytest.mark.asyncio
async def test_send_multiple_args(mock_websocket, vertexai):
  session = live.AsyncSession(
      api_client=mock_api_client(vertexai=vertexai), websocket=mock_websocket
  )
  with pytest.raises(ValueError, match='.*one argument.*'):
    await session.send_realtime_input(
        text='Hello?', activity_start=types.ActivityStart()
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Tests t_bytes methods in the _transformers module."""

import os

import PIL.Image
import pytest

from ... import _transformers as t
from ... import types


def test_blob_dict():
  blob = t.t_blob({
      'data': bytes([0, 0, 0, 0, 0, 0]),
      'mime_type': 'audio/pcm',
  })
  assert blob.data == bytes([0, 0, 0, 0, 0, 0])
  assert blob.mime_type == 'audio/pcm'


def test_blob():
  blob = t.t_blob(
      types.Blob(data=bytes([0, 0, 0, 0, 0, 0]), mime_type='audio/pcm')
  )
  assert blob.data == bytes([0, 0, 0, 0, 0, 0])
  assert blob.mime_type == 'audio/pcm'


def test_image(image_jpeg):
  blob = t.t_blob(image_jpeg)
  assert blob.data[6:10] == b'JFIF'
  assert blob.mime_type == 'image/jpeg'

  round_trip_image = blob.as_image()._pil_image
  assert round_trip_image is not None
  assert round_trip_image.size == image_jpeg.size
  assert round_trip_image.mode == image_jpeg.mode
  assert round_trip_image.format == image_jpeg.format


def test_not_image():
  blob = types.Blob(data=bytes([0, 0, 0, 0, 0, 0]), mime_type='audio/pcm')
  assert blob.as_image() is None


def test_part_image(image_jpeg):
  part = t.t_part(image_jpeg)
  assert part.inline_data.data[6:10] == b'JFIF'
  assert part.inline_data.mime_type == 'image/jpeg'

  round_trip_image = part.as_image()._pil_image
  assert round_trip_image is not None
  assert round_trip_image.size == image_jpeg.size
  assert round_trip_image.mode == image_jpeg.mode
  assert round_trip_image.format == image_jpeg.format


def test_part_not_image():
  part = t.t_part('hello world')
  assert part.as_image() is None


def test_pil_to_blob_with_memory_pil_image():
  img = PIL.Image.new('RGB', (1, 1), color='red')
  blob = t.pil_to_blob(img)
  assert blob.mime_type == 'image/png'
  assert blob.data and len(blob.data) == 69
  assert blob.data[0:4] == b'\x89PNG'
